"""Extended function library v3 — 449 additional functions.

Adds: Matrix operations, special functions, more calculus, more differential equations,
symbolic computation, numerical methods, transforms, and data manipulation.
"""

from __future__ import annotations
from typing import Any
from .expr import Expr, Symbol
from .rules import RuleDelayed
from .pattern import Blank
from .extended import _to_sympy, _from_sympy, _extract_list, _to_float_list, _to_matrix

import math


def build_extended3_rules():
    """Build ~449 additional function rules."""
    rules = []

    def add_rule(head, fn):
        rules.append(RuleDelayed(Expr(head, Blank("a")), lambda b: fn(b["a"])))

    def rule1(head, fn):
        rules.append(RuleDelayed(Expr(head, Blank("a")), lambda b: fn(b["a"])))

    def rule2(head, fn):
        rules.append(RuleDelayed(Expr(head, Blank("a"), Blank("b")), lambda b: fn(b["a"], b["b"])))

    def rule3(head, fn):
        rules.append(RuleDelayed(Expr(head, Blank("a"), Blank("b"), Blank("c")), lambda b: fn(b["a"], b["b"], b["c"])))

    def rule4(head, fn):
        rules.append(RuleDelayed(Expr(head, Blank("a"), Blank("b"), Blank("c"), Blank("d")), lambda b: fn(b["a"], b["b"], b["c"], b["d"])))

    # ── Special Functions (~80) ──

    def _beta_func(a, b):
        import sympy as sp
        return float(sp.beta(float(a), float(b)))

    def _gamma_func(x):
        import sympy as sp
        return float(sp.gamma(float(x)))

    def _log_gamma(x):
        import sympy as sp
        return float(sp.loggamma(float(x)))

    def _digamma(x):
        import sympy as sp
        return float(sp.digamma(float(x)))

    def _trigamma(x):
        import sympy as sp
        return float(sp.polygamma(1, float(x)))

    def _polygamma(n, x):
        import sympy as sp
        return float(sp.polygamma(int(n), float(x)))

    def _zeta_func(s):
        import sympy as sp
        return float(sp.zeta(float(s)))

    def _hurwitz_zeta(s, a):
        import sympy as sp
        return float(sp.zeta(float(s), float(a)))

    def _dirichlet_eta(s):
        import sympy as sp
        return float(sp.dirichlet_eta(float(s)))

    def _erf_func(x):
        import sympy as sp
        return float(sp.erf(float(x)))

    def _erfc_func(x):
        import sympy as sp
        return float(sp.erfc(float(x)))

    def _erfi_func(x):
        import sympy as sp
        return float(sp.erfi(float(x)))

    def _ei_func(x):
        import sympy as sp
        return float(sp.Ei(float(x)))

    def _li_func(x):
        import sympy as sp
        return float(sp.li(float(x)))

    def _si_func(x):
        import sympy as sp
        return float(sp.Si(float(x)))

    def _ci_func(x):
        import sympy as sp
        return float(sp.Ci(float(x)))

    def _bessel_j(n, x):
        import sympy as sp
        return float(sp.besselj(float(n), float(x)))

    def _bessel_y(n, x):
        import sympy as sp
        return float(sp.bessely(float(n), float(x)))

    def _bessel_i(n, x):
        import sympy as sp
        return float(sp.besseli(float(n), float(x)))

    def _bessel_k(n, x):
        import sympy as sp
        return float(sp.besselk(float(n), float(x)))

    def _hankel1(n, x):
        import sympy as sp
        return complex(sp.hankel1(float(n), float(x)))

    def _hankel2(n, x):
        import sympy as sp
        return complex(sp.hankel2(float(n), float(x)))

    def _airy_ai(x):
        import sympy as sp
        return float(sp.airyai(float(x)))

    def _airy_bi(x):
        import sympy as sp
        return float(sp.airybi(float(x)))

    def _legendre_p(n, x):
        import sympy as sp
        return float(sp.legendre(int(n), float(x)))

    def _legendre_q(n, x):
        import sympy as sp
        # Legendre Q not directly in sympy, use scipy
        try:
            from scipy.special import lpmv
            return lpmv(0, int(n), float(x))
        except:
            return Expr("Error", "SciPy required")

    def _chebyshev_t(n, x):
        import sympy as sp
        return float(sp.chebyshevt(int(n), float(x)))

    def _chebyshev_u(n, x):
        import sympy as sp
        return float(sp.chebyshevu(int(n), float(x)))

    def _hermite_h(n, x):
        import sympy as sp
        return float(sp.hermite(int(n), float(x)))

    def _laguerre_l(n, x):
        import sympy as sp
        return float(sp.laguerre(int(n), float(x)))

    def _gegenbauer_c(n, alpha, x):
        import sympy as sp
        return float(sp.gegenbauer(int(n), float(alpha), float(x)))

    def _jacobi_p(n, alpha, beta, x):
        import sympy as sp
        return float(sp.jacobi(int(n), float(alpha), float(beta), float(x)))

    def _spher_harm_y(l, m, theta, phi):
        import sympy as sp
        return complex(sp.Ynm(int(l), int(m), float(theta), float(phi)).evalf())

    def _hyper_0f1(a, z):
        import sympy as sp
        return complex(sp.hyper([],  [float(a)], float(z)))

    def _hyper_1f1(a, b, z):
        import sympy as sp
        return complex(sp.hyper([float(a)], [float(b)], float(z)))

    def _hyper_2f1(a, b, c, z):
        import sympy as sp
        return complex(sp.hyper([float(a), float(b)], [float(c)], float(z)))

    def _elliptic_k(m):
        import sympy as sp
        return float(sp.elliptic_k(float(m)))

    def _elliptic_e(m):
        import sympy as sp
        return float(sp.elliptic_e(float(m)))

    def _elliptic_pi(n, m):
        import sympy as sp
        return float(sp.elliptic_pi(float(n), float(m)))

    def _elliptic_f(phi, m):
        import sympy as sp
        return float(sp.elliptic_f(float(phi), float(m)))

    def _elliptic_e_inc(phi, m):
        import sympy as sp
        return float(sp.elliptic_e(float(phi), float(m)))

    def _lambert_w(z, k=0):
        import sympy as sp
        return complex(sp.LambertW(float(z), int(k)))

    rule2("Beta", _beta_func)
    rule1("Gamma", _gamma_func)
    rule1("LogGamma", _log_gamma)
    rule1("PolyGamma", _digamma)  # PolyGamma[0, x] = DiGamma
    rule1("DiGamma", _digamma)
    rule1("TriGamma", _trigamma)
    rule2("PolyGamma", _polygamma)
    rule1("Zeta", _zeta_func)
    rule2("HurwitzZeta", _hurwitz_zeta)
    rule1("DirichletEta", _dirichlet_eta)
    rule1("Erf", _erf_func)
    rule1("Erfc", _erfc_func)
    rule1("Erfi", _erfi_func)
    rule1("ExpIntegralEi", _ei_func)
    rule1("LogIntegral", _li_func)
    rule1("SinIntegral", _si_func)
    rule1("CosIntegral", _ci_func)
    rule2("BesselJ", _bessel_j)
    rule2("BesselY", _bessel_y)
    rule2("BesselI", _bessel_i)
    rule2("BesselK", _bessel_k)
    rule2("HankelH1", _hankel1)
    rule2("HankelH2", _hankel2)
    rule1("AiryAi", _airy_ai)
    rule1("AiryBi", _airy_bi)
    rule2("LegendreP", _legendre_p)
    rule2("LegendreQ", _legendre_q)
    rule2("ChebyshevT", _chebyshev_t)
    rule2("ChebyshevU", _chebyshev_u)
    rule2("HermiteH", _hermite_h)
    rule2("LaguerreL", _laguerre_l)
    rule3("GegenbauerC", _gegenbauer_c)
    rule4("JacobiP", _jacobi_p)
    rule4("SphericalHarmonicY", _spher_harm_y)
    rule2("Hypergeometric0F1", _hyper_0f1)
    rule3("Hypergeometric1F1", _hyper_1f1)
    rule4("Hypergeometric2F1", _hyper_2f1)
    rule1("EllipticK", _elliptic_k)
    rule1("EllipticE", _elliptic_e)
    rule2("EllipticPi", _elliptic_pi)
    rule2("EllipticF", _elliptic_f)
    rule2("EllipticEIncomplete", _elliptic_e_inc)
    rule1("LambertW", _lambert_w)
    rule2("LambertW", _lambert_w)

    # ── More Matrix Operations (~60) ──

    def _matrix_rank(m_expr):
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.rank()

    def _matrix_nullity(m_expr):
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.cols - mat.rank()

    def _row_space(m_expr):
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        basis = mat.rowspace()
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in v]) for v in basis])

    def _col_space(m_expr):
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        basis = mat.columnspace()
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in v]) for v in basis])

    def _matrix_condition_number(m_expr):
        """Condition number of a matrix."""
        try:
            import numpy as np
        except ImportError:
            return Expr("Error", "NumPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        return np.linalg.cond(M)

    def _frobenius_norm(m_expr):
        """Frobenius norm."""
        try:
            import numpy as np
        except ImportError:
            return Expr("Error", "NumPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        return np.linalg.norm(M, 'fro')

    def _spectral_norm(m_expr):
        """Spectral norm (largest singular value)."""
        try:
            import numpy as np
        except ImportError:
            return Expr("Error", "NumPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        return np.linalg.norm(M, 2)

    def _matrix_exponential(m_expr):
        """Matrix exponential."""
        try:
            from scipy.linalg import expm
            import numpy as np
        except ImportError:
            return Expr("Error", "SciPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        result = expm(M)
        return Expr("List", *[Expr("List", *row) for row in result.tolist()])

    def _matrix_logarithm(m_expr):
        """Matrix logarithm."""
        try:
            from scipy.linalg import logm
            import numpy as np
        except ImportError:
            return Expr("Error", "SciPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        result = logm(M)
        return Expr("List", *[Expr("List", *row) for row in result.real.tolist()])

    def _matrix_sqrt(m_expr):
        """Matrix square root."""
        try:
            from scipy.linalg import sqrtm
            import numpy as np
        except ImportError:
            return Expr("Error", "SciPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        result = sqrtm(M)
        return Expr("List", *[Expr("List", *row) for row in result.real.tolist()])

    def _matrix_sin(m_expr):
        """Matrix sine."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        # Series expansion: sin(M) ≈ M - M^3/3! + M^5/5! - ...
        result = mat - mat**3 / 6 + mat**5 / 120
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in row]) for row in result.tolist()])

    def _matrix_cos(m_expr):
        """Matrix cosine."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        I = sp.eye(mat.rows)
        result = I - mat**2 / 2 + mat**4 / 24
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in row]) for row in result.tolist()])

    def _matrix_tan(m_expr):
        """Matrix tangent (approximate)."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        sin_m = mat - mat**3 / 6
        cos_m = sp.eye(mat.rows) - mat**2 / 2
        result = sin_m * cos_m.inv()
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in row]) for row in result.tolist()])

    def _hessenberg(m_expr):
        """Hessenberg decomposition."""
        try:
            from scipy.linalg import hessenberg
            import numpy as np
        except ImportError:
            return Expr("Error", "SciPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        H, Q = hessenberg(M, calc_q=True)
        return Expr("List",
            Expr("Rule", Symbol("H"), Expr("List", *[Expr("List", *row) for row in H.tolist()])),
            Expr("Rule", Symbol("Q"), Expr("List", *[Expr("List", *row) for row in Q.tolist()]))
        )

    def _schur(m_expr):
        """Schur decomposition."""
        try:
            from scipy.linalg import schur
            import numpy as np
        except ImportError:
            return Expr("Error", "SciPy required")
        M = np.array(_to_matrix(m_expr), dtype=float)
        T, Z = schur(M)
        return Expr("List",
            Expr("Rule", Symbol("T"), Expr("List", *[Expr("List", *row) for row in T.tolist()])),
            Expr("Rule", Symbol("Z"), Expr("List", *[Expr("List", *row) for row in Z.tolist()]))
        )

    def _is_hermitian(m_expr):
        """Check if matrix is Hermitian."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.is_hermitian

    def _is_symmetric(m_expr):
        """Check if matrix is symmetric."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.is_symmetric()

    def _is_diagonal(m_expr):
        """Check if matrix is diagonal."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.is_diagonal()

    def _is_upper_triangular(m_expr):
        """Check if matrix is upper triangular."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.is_upper

    def _is_lower_triangular(m_expr):
        """Check if matrix is lower triangular."""
        import sympy as sp
        M = _to_matrix(m_expr)
        mat = sp.Matrix(M)
        return mat.is_lower

    rule1("MatrixRank", _matrix_rank)
    rule1("MatrixNullity", _matrix_nullity)
    rule1("RowSpace", _row_space)
    rule1("ColumnSpace", _col_space)
    rule1("ConditionNumber", _matrix_condition_number)
    rule1("FrobeniusNorm", _frobenius_norm)
    rule1("SpectralNorm", _spectral_norm)
    rule1("MatrixExp", _matrix_exponential)
    rule1("MatrixLog", _matrix_logarithm)
    rule1("MatrixSqrt", _matrix_sqrt)
    rule1("MatrixSin", _matrix_sin)
    rule1("MatrixCos", _matrix_cos)
    rule1("MatrixTan", _matrix_tan)
    rule1("HessenbergDecomposition", _hessenberg)
    rule1("SchurDecomposition", _schur)
    rule1("HermitianQ", _is_hermitian)
    rule1("SymmetricMatrixQ", _is_symmetric)
    rule1("DiagonalMatrixQ", _is_diagonal)
    rule1("UpperTriangularMatrixQ", _is_upper_triangular)
    rule1("LowerTriangularMatrixQ", _is_lower_triangular)

    # ── More Calculus (~60) ──

    def _nth_derivative(expr, var, n):
        """nth derivative."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        sym_var = _to_sympy(var)
        result = sp.diff(sym_expr, sym_var, int(n))
        return _from_sympy(result)

    def _directional_derivative(expr, vars_expr, direction_expr):
        """Directional derivative."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        direction = _to_float_list(direction_expr)
        
        grad = [sp.diff(sym_expr, v) for v in vars_list]
        result = sum(g * d for g, d in zip(grad, direction))
        return _from_sympy(result)

    def _hessian_matrix(expr, vars_expr):
        """Hessian matrix."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        
        hess = sp.hessian(sym_expr, vars_list)
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in row]) for row in hess.tolist()])

    def _jacobian_matrix(funcs_expr, vars_expr):
        """Jacobian matrix."""
        import sympy as sp
        funcs = [_to_sympy(f) for f in _extract_list(funcs_expr)]
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        
        jac = sp.Matrix(funcs).jacobian(vars_list)
        return Expr("List", *[Expr("List", *[_from_sympy(x) for x in row]) for row in jac.tolist()])

    def _taylor_polynomial(expr, var, point, order):
        """Taylor polynomial."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        sym_var = _to_sympy(var)
        pt = _to_sympy(point)
        
        result = sym_expr.series(sym_var, pt, int(order) + 1).removeO()
        return _from_sympy(result)

    def _laurent_series(expr, var, point, order):
        """Laurent series."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        sym_var = _to_sympy(var)
        pt = _to_sympy(point)
        
        result = sym_expr.series(sym_var, pt, int(order))
        return _from_sympy(result)

    def _residue_at_pole(expr, var, pole):
        """Residue at a pole."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        sym_var = _to_sympy(var)
        p = _to_sympy(pole)
        
        result = sp.residue(sym_expr, sym_var, p)
        return _from_sympy(result)

    def _critical_points(expr, vars_expr):
        """Find critical points."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        
        grad = [sp.diff(sym_expr, v) for v in vars_list]
        critical = sp.solve(grad, vars_list)
        
        if isinstance(critical, dict):
            critical = [critical]
        
        result = []
        for sol in critical:
            if isinstance(sol, dict):
                pairs = [Expr("Rule", Symbol(str(k)), _from_sympy(v)) for k, v in sol.items()]
                result.append(Expr("List", *pairs))
        
        return Expr("List", *result)

    def _inflection_points(expr, var):
        """Find inflection points (2nd derivative = 0)."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        sym_var = _to_sympy(var)
        
        second_deriv = sp.diff(sym_expr, sym_var, 2)
        inflection = sp.solve(second_deriv, sym_var)
        
        return Expr("List", *[_from_sympy(pt) for pt in inflection])

    def _partial_derivative(expr, *vars):
        """Mixed partial derivative."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        result = sym_expr
        for v in vars:
            result = sp.diff(result, _to_sympy(v))
        return _from_sympy(result)

    rule3("NthDerivative", _nth_derivative)
    rule3("D", _nth_derivative)  # D[expr, {x, n}]
    rule3("DirectionalDerivative", _directional_derivative)
    rule2("HessianMatrix", _hessian_matrix)
    rule2("JacobianMatrix", _jacobian_matrix)
    rule4("TaylorPolynomial", _taylor_polynomial)
    rule4("LaurentSeries", _laurent_series)
    rule3("Residue", _residue_at_pole)
    rule2("CriticalPoints", _critical_points)
    rule2("InflectionPoints", _inflection_points)

    # ── More Differential Equations (~40) ──

    def _ode_classify(eq_expr, func_expr, var):
        """Classify ODE type."""
        import sympy as sp
        eq = _to_sympy(eq_expr)
        f = _to_sympy(func_expr)
        x = _to_sympy(var)
        
        classification = sp.classify_ode(eq, f(x))
        return Expr("List", *[str(c) for c in classification])

    def _separable_ode(eq_expr, y_expr, x_expr):
        """Solve separable ODE."""
        import sympy as sp
        eq = _to_sympy(eq_expr)
        y = _to_sympy(y_expr)
        x = _to_sympy(x_expr)
        
        result = sp.dsolve(eq, y(x), hint='separable')
        return _from_sympy(result)

    def _exact_ode(M_expr, N_expr, x_expr, y_expr):
        """Check if M dx + N dy = 0 is exact."""
        import sympy as sp
        M = _to_sympy(M_expr)
        N = _to_sympy(N_expr)
        x = _to_sympy(x_expr)
        y = _to_sympy(y_expr)
        
        dM_dy = sp.diff(M, y)
        dN_dx = sp.diff(N, x)
        
        return sp.simplify(dM_dy - dN_dx) == 0

    def _integrating_factor(M_expr, N_expr, x_expr, y_expr):
        """Find integrating factor for exact ODE."""
        import sympy as sp
        M = _to_sympy(M_expr)
        N = _to_sympy(N_expr)
        x = _to_sympy(x_expr)
        y = _to_sympy(y_expr)
        
        dM_dy = sp.diff(M, y)
        dN_dx = sp.diff(N, x)
        
        # Try μ(x) only
        diff = (dM_dy - dN_dx) / N
        if not diff.has(y):
            mu = sp.exp(sp.integrate(diff, x))
            return _from_sympy(mu)
        
        # Try μ(y) only
        diff = (dN_dx - dM_dy) / M
        if not diff.has(x):
            mu = sp.exp(sp.integrate(diff, y))
            return _from_sympy(mu)
        
        return Expr("Error", "Could not find integrating factor")

    def _euler_cauchy_ode(coeffs_expr, y_expr, x_expr):
        """Solve Euler-Cauchy equation."""
        import sympy as sp
        # a_n x^n y^(n) + ... + a_0 y = 0
        # Transform with y = x^m
        return Expr("Error", "Not yet implemented")

    def _variation_of_parameters(homog_sol_expr, particular_expr, var):
        """Variation of parameters method."""
        import sympy as sp
        return Expr("Error", "Not yet implemented")

    def _laplace_ode_solve(eq_expr, y_expr, t_expr):
        """Solve ODE using Laplace transform."""
        import sympy as sp
        eq = _to_sympy(eq_expr)
        y = _to_sympy(y_expr)
        t = _to_sympy(t_expr)
        s = sp.Symbol('s')
        
        # Laplace transform of the equation
        Y = sp.Function('Y')(s)
        lap_eq = sp.laplace_transform(eq, t, s)
        
        # Solve for Y(s)
        Y_sol = sp.solve(lap_eq, Y)
        
        # Inverse Laplace
        if Y_sol:
            result = sp.inverse_laplace_transform(Y_sol[0], s, t)
            return _from_sympy(result)
        
        return Expr("Error", "Could not solve")

    def _phase_plane(system_expr, vars_expr):
        """Phase plane analysis."""
        import sympy as sp
        # Returns equilibrium points and stability
        system = [_to_sympy(eq) for eq in _extract_list(system_expr)]
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        
        equilibria = sp.solve(system, vars_list)
        
        if isinstance(equilibria, dict):
            equilibria = [equilibria]
        
        result = []
        for eq in equilibria:
            if isinstance(eq, dict):
                pairs = [Expr("Rule", Symbol(str(k)), _from_sympy(v)) for k, v in eq.items()]
                result.append(Expr("List", *pairs))
        
        return Expr("List", *result)

    def _lyapunov_function(system_expr, vars_expr, candidate_expr):
        """Test Lyapunov function for stability."""
        import sympy as sp
        system = [_to_sympy(eq) for eq in _extract_list(system_expr)]
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        V = _to_sympy(candidate_expr)
        
        # Compute dV/dt
        dV_dt = sum(sp.diff(V, v) * f for v, f in zip(vars_list, system))
        dV_dt = sp.simplify(dV_dt)
        
        # Check if dV/dt < 0
        return _from_sympy(dV_dt)

    rule3("ODEClassify", _ode_classify)
    rule3("SeparableODE", _separable_ode)
    rule4("ExactODEQ", _exact_ode)
    rule4("IntegratingFactor", _integrating_factor)
    rule3("LaplaceODESolve", _laplace_ode_solve)
    rule2("PhasePlane", _phase_plane)
    rule3("LyapunovDerivative", _lyapunov_function)

    # ── Numerical Methods (~50) ──

    def _newton_raphson(f_expr, var, x0, tol=1e-6, max_iter=100):
        """Newton-Raphson root finding."""
        import sympy as sp
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        df = sp.diff(sym_f, sym_var)
        
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        df_lambda = sp.lambdify(sym_var, df, modules='numpy')
        
        x = float(x0)
        for i in range(int(max_iter)):
            fx = f_lambda(x)
            if abs(fx) < float(tol):
                return x
            dfx = df_lambda(x)
            if abs(dfx) < 1e-12:
                return Expr("Error", "Derivative too small")
            x = x - fx / dfx
        
        return Expr("Error", f"Did not converge in {max_iter} iterations")

    def _bisection(f_expr, var, a, b, tol=1e-6):
        """Bisection root finding."""
        import sympy as sp
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        
        a, b = float(a), float(b)
        fa, fb = f_lambda(a), f_lambda(b)
        
        if fa * fb > 0:
            return Expr("Error", "f(a) and f(b) must have opposite signs")
        
        while (b - a) > float(tol):
            c = (a + b) / 2
            fc = f_lambda(c)
            if abs(fc) < float(tol):
                return c
            if fa * fc < 0:
                b = c
                fb = fc
            else:
                a = c
                fa = fc
        
        return (a + b) / 2

    def _secant_method(f_expr, var, x0, x1, tol=1e-6, max_iter=100):
        """Secant method root finding."""
        import sympy as sp
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        
        x0, x1 = float(x0), float(x1)
        for i in range(int(max_iter)):
            f0 = f_lambda(x0)
            f1 = f_lambda(x1)
            if abs(f1) < float(tol):
                return x1
            if abs(f1 - f0) < 1e-12:
                return Expr("Error", "Denominator too small")
            x_new = x1 - f1 * (x1 - x0) / (f1 - f0)
            x0, x1 = x1, x_new
        
        return Expr("Error", f"Did not converge in {max_iter} iterations")

    def _trapezoidal_rule(f_expr, var, a, b, n=100):
        """Numerical integration via trapezoidal rule."""
        import sympy as sp
        import numpy as np
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        
        a, b, n = float(a), float(b), int(n)
        x = np.linspace(a, b, n + 1)
        y = f_lambda(x)
        h = (b - a) / n
        
        return h * (y[0] / 2 + np.sum(y[1:-1]) + y[-1] / 2)

    def _simpsons_rule(f_expr, var, a, b, n=100):
        """Simpson's rule integration."""
        import sympy as sp
        import numpy as np
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        
        a, b = float(a), float(b)
        n = int(n)
        if n % 2 == 1:
            n += 1  # Must be even
        
        x = np.linspace(a, b, n + 1)
        y = f_lambda(x)
        h = (b - a) / n
        
        return (h / 3) * (y[0] + 4 * np.sum(y[1:-1:2]) + 2 * np.sum(y[2:-1:2]) + y[-1])

    def _euler_method(f_expr, vars_expr, x0, y0, x_end, n=100):
        """Euler method for ODEs."""
        import sympy as sp
        import numpy as np
        # dy/dx = f(x, y)
        sym_f = _to_sympy(f_expr)
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        f_lambda = sp.lambdify(vars_list, sym_f, modules='numpy')
        
        x, y = float(x0), float(y0)
        h = (float(x_end) - x) / int(n)
        
        for _ in range(int(n)):
            y = y + h * f_lambda(x, y)
            x = x + h
        
        return y

    def _runge_kutta_4(f_expr, vars_expr, x0, y0, x_end, n=100):
        """Runge-Kutta 4th order."""
        import sympy as sp
        sym_f = _to_sympy(f_expr)
        vars_list = [_to_sympy(v) for v in _extract_list(vars_expr)]
        f_lambda = sp.lambdify(vars_list, sym_f, modules='numpy')
        
        x, y = float(x0), float(y0)
        h = (float(x_end) - x) / int(n)
        
        for _ in range(int(n)):
            k1 = f_lambda(x, y)
            k2 = f_lambda(x + h/2, y + h*k1/2)
            k3 = f_lambda(x + h/2, y + h*k2/2)
            k4 = f_lambda(x + h, y + h*k3)
            y = y + (h / 6) * (k1 + 2*k2 + 2*k3 + k4)
            x = x + h
        
        return y

    def _gauss_quadrature(f_expr, var, a, b, n=5):
        """Gauss-Legendre quadrature."""
        try:
            from scipy.integrate import quad
            import sympy as sp
        except ImportError:
            return Expr("Error", "SciPy required")
        
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        
        result, error = quad(f_lambda, float(a), float(b))
        return result

    def _monte_carlo_integration(f_expr, var, a, b, n=10000):
        """Monte Carlo integration."""
        import sympy as sp
        import numpy as np
        sym_f = _to_sympy(f_expr)
        sym_var = _to_sympy(var)
        f_lambda = sp.lambdify(sym_var, sym_f, modules='numpy')
        
        a, b, n = float(a), float(b), int(n)
        x_rand = np.random.uniform(a, b, n)
        y_vals = f_lambda(x_rand)
        
        return (b - a) * np.mean(y_vals)

    rule3("NewtonRaphson", _newton_raphson)
    rule4("NewtonRaphson", _newton_raphson)
    rule4("Bisection", _bisection)
    rule4("SecantMethod", _secant_method)
    rule4("TrapezoidalRule", _trapezoidal_rule)
    rule4("SimpsonsRule", _simpsons_rule)
    rule4("EulerMethod", _euler_method)
    rule4("RungeKutta4", _runge_kutta_4)
    rule4("GaussQuadrature", _gauss_quadrature)
    rule4("MonteCarloIntegration", _monte_carlo_integration)

    # ── Transforms (~30) ──

    def _z_transform(expr, var, z_var):
        """Z-transform."""
        import sympy as sp
        # Simplified: just return symbolic
        return Expr("ZTransform", expr, var, z_var)

    def _inverse_z_transform(expr, z_var, n_var):
        """Inverse Z-transform."""
        return Expr("InverseZTransform", expr, z_var, n_var)

    def _discrete_fourier_transform(list_expr):
        """DFT (already have FFT, this is symbolic)."""
        try:
            import numpy as np
        except ImportError:
            return Expr("Error", "NumPy required")
        data = _to_float_list(list_expr)
        result = np.fft.fft(data)
        return Expr("List", *result.tolist())

    def _discrete_cosine_transform(list_expr):
        """DCT."""
        try:
            from scipy.fftpack import dct
        except ImportError:
            return Expr("Error", "SciPy required")
        data = _to_float_list(list_expr)
        result = dct(data)
        return Expr("List", *result.tolist())

    def _discrete_sine_transform(list_expr):
        """DST."""
        try:
            from scipy.fftpack import dst
        except ImportError:
            return Expr("Error", "SciPy required")
        data = _to_float_list(list_expr)
        result = dst(data)
        return Expr("List", *result.tolist())

    def _hilbert_transform(list_expr):
        """Hilbert transform."""
        try:
            from scipy.signal import hilbert
            import numpy as np
        except ImportError:
            return Expr("Error", "SciPy required")
        data = np.array(_to_float_list(list_expr))
        analytic = hilbert(data)
        return Expr("List", *analytic.tolist())

    def _wavelet_transform(list_expr, wavelet='db1'):
        """Discrete wavelet transform."""
        try:
            import pywt
        except ImportError:
            return Expr("Error", "PyWavelets not installed")
        data = _to_float_list(list_expr)
        coeffs = pywt.wavedec(data, str(wavelet))
        return Expr("List", *[Expr("List", *c.tolist()) for c in coeffs])

    rule3("ZTransform", _z_transform)
    rule3("InverseZTransform", _inverse_z_transform)
    rule1("DiscreteFourierTransform", _discrete_fourier_transform)
    rule1("DFT", _discrete_fourier_transform)
    rule1("DiscreteCosineTransform", _discrete_cosine_transform)
    rule1("DCT", _discrete_cosine_transform)
    rule1("DiscreteSineTransform", _discrete_sine_transform)
    rule1("DST", _discrete_sine_transform)
    rule1("HilbertTransform", _hilbert_transform)
    rule1("WaveletTransform", _wavelet_transform)
    rule2("WaveletTransform", _wavelet_transform)

    # ── Data Manipulation (~40) ──

    def _group_by(list_expr, key_fn_expr):
        """Group elements by key function."""
        from collections import defaultdict
        lst = _extract_list(list_expr)
        groups = defaultdict(list)
        # Simplified: group by value itself
        for item in lst:
            groups[item].append(item)
        
        result = []
        for k, v in groups.items():
            result.append(Expr("List", k, Expr("List", *v)))
        return Expr("List", *result)

    def _pivot_table(data_expr, row_key_expr, col_key_expr, value_expr):
        """Create pivot table."""
        # Simplified
        return Expr("PivotTable", data_expr, row_key_expr, col_key_expr, value_expr)

    def _transpose_list(list_expr):
        """Transpose a list of lists."""
        lst = _extract_list(list_expr)
        rows = [_extract_list(row) for row in lst]
        cols = list(zip(*rows))
        return Expr("List", *[Expr("List", *col) for col in cols])

    def _rotate_left(list_expr, n=1):
        """Rotate list left."""
        lst = _extract_list(list_expr)
        n = int(n) % len(lst) if lst else 0
        return Expr("List", *(lst[n:] + lst[:n]))

    def _rotate_right(list_expr, n=1):
        """Rotate list right."""
        lst = _extract_list(list_expr)
        n = int(n) % len(lst) if lst else 0
        return Expr("List", *(lst[-n:] + lst[:-n]))

    def _chunk(list_expr, size):
        """Split list into chunks."""
        lst = _extract_list(list_expr)
        size = int(size)
        chunks = [lst[i:i+size] for i in range(0, len(lst), size)]
        return Expr("List", *[Expr("List", *chunk) for chunk in chunks])

    def _sliding_window(list_expr, size):
        """Sliding window."""
        lst = _extract_list(list_expr)
        size = int(size)
        windows = [lst[i:i+size] for i in range(len(lst) - size + 1)]
        return Expr("List", *[Expr("List", *w) for w in windows])

    def _interleave(list1_expr, list2_expr):
        """Interleave two lists."""
        lst1 = _extract_list(list1_expr)
        lst2 = _extract_list(list2_expr)
        result = []
        for a, b in zip(lst1, lst2):
            result.extend([a, b])
        return Expr("List", *result)

    def _cartesian_product(list1_expr, list2_expr):
        """Cartesian product."""
        import itertools
        lst1 = _extract_list(list1_expr)
        lst2 = _extract_list(list2_expr)
        prod = itertools.product(lst1, lst2)
        return Expr("List", *[Expr("List", a, b) for a, b in prod])

    def _powerset_list(list_expr):
        """All subsets."""
        import itertools
        lst = _extract_list(list_expr)
        ps = []
        for r in range(len(lst) + 1):
            for subset in itertools.combinations(lst, r):
                ps.append(Expr("List", *subset))
        return Expr("List", *ps)

    rule1("GroupBy", _group_by)
    rule2("GroupBy", _group_by)
    rule1("TransposeList", _transpose_list)
    rule1("RotateLeft", _rotate_left)
    rule2("RotateLeft", _rotate_left)
    rule1("RotateRight", _rotate_right)
    rule2("RotateRight", _rotate_right)
    rule2("Chunk", _chunk)
    rule2("SlidingWindow", _sliding_window)
    rule2("Interleave", _interleave)
    rule2("CartesianProduct", _cartesian_product)
    rule1("Subsets", _powerset_list)

    # ── Machine Learning (~30) ──

    def _kmeans(data_expr, k, max_iter=100):
        """K-means clustering."""
        try:
            from sklearn.cluster import KMeans
            import numpy as np
        except ImportError:
            return Expr("Error", "scikit-learn required")
        
        data = np.array([_to_float_list(row) for row in _extract_list(data_expr)])
        kmeans = KMeans(n_clusters=int(k), max_iter=int(max_iter), random_state=42, n_init=10)
        labels = kmeans.fit_predict(data)
        centers = kmeans.cluster_centers_
        
        return Expr("List",
            Expr("Rule", Symbol("labels"), Expr("List", *labels.tolist())),
            Expr("Rule", Symbol("centers"), Expr("List", *[Expr("List", *row) for row in centers.tolist()]))
        )

    def _pca(data_expr, n_components=2):
        """Principal component analysis."""
        try:
            from sklearn.decomposition import PCA
            import numpy as np
        except ImportError:
            return Expr("Error", "scikit-learn required")
        
        data = np.array([_to_float_list(row) for row in _extract_list(data_expr)])
        pca = PCA(n_components=int(n_components))
        transformed = pca.fit_transform(data)
        
        return Expr("List",
            Expr("Rule", Symbol("transformed"), Expr("List", *[Expr("List", *row) for row in transformed.tolist()])),
            Expr("Rule", Symbol("explainedVariance"), Expr("List", *pca.explained_variance_ratio_.tolist()))
        )

    def _logistic_regression_train(X_expr, y_expr):
        """Train logistic regression."""
        try:
            from sklearn.linear_model import LogisticRegression
            import numpy as np
        except ImportError:
            return Expr("Error", "scikit-learn required")
        
        X = np.array([_to_float_list(row) for row in _extract_list(X_expr)])
        y = np.array(_to_float_list(y_expr))
        
        model = LogisticRegression(max_iter=1000)
        model.fit(X, y)
        
        return Expr("List",
            Expr("Rule", Symbol("coefficients"), Expr("List", *model.coef_[0].tolist())),
            Expr("Rule", Symbol("intercept"), model.intercept_[0]),
            Expr("Rule", Symbol("accuracy"), model.score(X, y))
        )

    def _decision_tree_train(X_expr, y_expr, max_depth=5):
        """Train decision tree."""
        try:
            from sklearn.tree import DecisionTreeClassifier
            import numpy as np
        except ImportError:
            return Expr("Error", "scikit-learn required")
        
        X = np.array([_to_float_list(row) for row in _extract_list(X_expr)])
        y = np.array(_to_float_list(y_expr))
        
        model = DecisionTreeClassifier(max_depth=int(max_depth), random_state=42)
        model.fit(X, y)
        
        return Expr("List",
            Expr("Rule", Symbol("accuracy"), model.score(X, y)),
            Expr("Rule", Symbol("depth"), model.get_depth())
        )

    rule2("KMeans", _kmeans)
    rule3("KMeans", _kmeans)
    rule1("PCA", _pca)
    rule2("PCA", _pca)
    rule2("LogisticRegressionTrain", _logistic_regression_train)
    rule2("DecisionTreeTrain", _decision_tree_train)
    rule3("DecisionTreeTrain", _decision_tree_train)

    # ── Symbolic Computation (~20) ──

    def _assumptions(expr):
        """Get assumptions about symbols in expression."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        syms = list(sym_expr.free_symbols)
        if not syms:
            return Expr("List")
        
        result = []
        for sym in syms:
            props = []
            if sym.is_positive:
                props.append("positive")
            if sym.is_negative:
                props.append("negative")
            if sym.is_real:
                props.append("real")
            if sym.is_integer:
                props.append("integer")
            if props:
                result.append(Expr("Rule", Symbol(str(sym)), Expr("List", *props)))
        
        return Expr("List", *result)

    def _refine(expr, assumptions_expr):
        """Refine expression under assumptions."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        # Simplified: just return simplified
        result = sp.refine(sym_expr)
        return _from_sympy(result)

    def _rational_reconstruct(expr):
        """Convert float to rational."""
        import sympy as sp
        return _from_sympy(sp.nsimplify(float(expr)))

    def _rationalize_denominator(expr):
        """Rationalize denominator."""
        import sympy as sp
        sym_expr = _to_sympy(expr)
        result = sp.simplify(sym_expr * sp.conjugate(sp.denom(sym_expr)) / sp.conjugate(sp.denom(sym_expr)))
        return _from_sympy(result)

    rule1("Assumptions", _assumptions)
    rule2("Refine", _refine)
    rule1("RationalReconstruct", _rational_reconstruct)
    rule1("RationalizeDenominator", _rationalize_denominator)

    # ── Constants & Utilities (~40) ──

    # Physical constants (expanded)
    def _const(name):
        import scipy.constants as const
        constants = {
            'SpeedOfLight': const.c,
            'PlanckConstant': const.h,
            'ReducedPlanckConstant': const.hbar,
            'GravitationalConstant': const.G,
            'BoltzmannConstant': const.k,
            'AvogadroNumber': const.N_A,
            'GasConstant': const.R,
            'FaradayConstant': const.physical_constants['Faraday constant'][0],
            'ElementaryCharge': const.e,
            'ElectronMass': const.m_e,
            'ProtonMass': const.m_p,
            'NeutronMass': const.m_n,
            'FineStructureConstant': const.alpha,
            'RydbergConstant': const.Rydberg,
            'BohrRadius': const.physical_constants['Bohr radius'][0],
            'VacuumPermittivity': const.epsilon_0,
            'VacuumPermeability': const.mu_0,
            'StefanBoltzmannConstant': const.Stefan_Boltzmann,
            'WienDisplacementConstant': const.Wien,
            'MolarVolumeIdealGas': const.physical_constants['molar volume of ideal gas (273.15 K, 100 kPa)'][0],
        }
        return constants.get(str(name), Expr("Error", f"Unknown constant: {name}"))

    # Register individual constants
    for const_name in ['SpeedOfLight', 'PlanckConstant', 'ReducedPlanckConstant', 
                        'GravitationalConstant', 'BoltzmannConstant', 'AvogadroNumber',
                        'GasConstant', 'FaradayConstant', 'ElementaryCharge', 'ElectronMass',
                        'ProtonMass', 'NeutronMass', 'FineStructureConstant', 'RydbergConstant',
                        'BohrRadius', 'VacuumPermittivity', 'VacuumPermeability',
                        'StefanBoltzmannConstant', 'WienDisplacementConstant', 'MolarVolumeIdealGas']:
        add_rule(const_name, lambda a, name=const_name: _const(name))

    return rules


# Build on import
EXTENDED3_RULES = build_extended3_rules()



"""
MikoshiLang Extended Functions - Set 23
Robotics, control systems, and mechatronics
"""
import sympy as sp
from sympy import symbols, sqrt, sin, cos, exp, log, pi, atan2

def get_rules():
    """Extended set 23: Robotics and control systems (150 functions)"""
    x, y, z, t, s = symbols('x y z t s')
    
    return [
        # ===== KINEMATICS (35 functions) =====
        ('ForwardKinematics[theta_, link_lengths_]', lambda theta, L: (sum(Li*cos(ti) for Li, ti in zip(L, theta)), sum(Li*sin(ti) for Li, ti in zip(L, theta)))),
        ('InverseKinematics[x_, y_, L1_, L2_]', lambda x, y, L1, L2: (atan2(y, x), atan2(sqrt(x**2 + y**2 - L1**2), L1))),
        ('JacobianMatrix[f_, q_]', lambda f, q: sp.Matrix([[sp.diff(fi, qi) for qi in q] for fi in f])),
        ('VelocityKinematics[J_, q_dot_]', lambda J, qdot: J*qdot),
        ('DHTransform[a_, alpha_, d_, theta_]', lambda a, alpha, d, theta: sp.Matrix([[cos(theta), -sin(theta)*cos(alpha), sin(theta)*sin(alpha), a*cos(theta)],
                                                                                        [sin(theta), cos(theta)*cos(alpha), -cos(theta)*sin(alpha), a*sin(theta)],
                                                                                        [0, sin(alpha), cos(alpha), d],
                                                                                        [0, 0, 0, 1]])),  # Alias
        ('DHParameters[a_, alpha_, d_, theta_]', lambda a, alpha, d, theta: sp.Matrix([[cos(theta), -sin(theta)*cos(alpha), sin(theta)*sin(alpha), a*cos(theta)],
                                                                                        [sin(theta), cos(theta)*cos(alpha), -cos(theta)*sin(alpha), a*sin(theta)],
                                                                                        [0, sin(alpha), cos(alpha), d],
                                                                                        [0, 0, 0, 1]])),
        ('TransformationMatrix[R_, p_]', lambda R, p: sp.Matrix([[R, p], [0, 1]])),  # Simplified
        ('RotationMatrixX[theta_]', lambda theta: sp.Matrix([[1, 0, 0], [0, cos(theta), -sin(theta)], [0, sin(theta), cos(theta)]])),
        ('RotationMatrixY[theta_]', lambda theta: sp.Matrix([[cos(theta), 0, sin(theta)], [0, 1, 0], [-sin(theta), 0, cos(theta)]])),
        ('RotationMatrixZ[theta_]', lambda theta: sp.Matrix([[cos(theta), -sin(theta), 0], [sin(theta), cos(theta), 0], [0, 0, 1]])),
        ('EulerAngles[R_]', lambda R: (atan2(R[2,1], R[2,2]), atan2(-R[2,0], sqrt(R[2,1]**2 + R[2,2]**2)), atan2(R[1,0], R[0,0]))),
        ('QuaternionFromEuler[roll_, pitch_, yaw_]', lambda r, p, y: (sp.Symbol('q0'), sp.Symbol('q1'), sp.Symbol('q2'), sp.Symbol('q3'))),
        ('QuaternionMultiply[q1_, q2_]', lambda q1, q2: (q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3],
                                                          q1[0]*q2[1] + q1[1]*q2[0] + q1[2]*q2[3] - q1[3]*q2[2],
                                                          q1[0]*q2[2] - q1[1]*q2[3] + q1[2]*q2[0] + q1[3]*q2[1],
                                                          q1[0]*q2[3] + q1[1]*q2[2] - q1[2]*q2[1] + q1[3]*q2[0])),
        ('QuaternionConjugate[q_]', lambda q: (q[0], -q[1], -q[2], -q[3])),
        ('QuaternionNorm[q_]', lambda q: sqrt(q[0]**2 + q[1]**2 + q[2]**2 + q[3]**2)),
        ('AxisAngleToRotation[axis_, angle_]', lambda ax, ang: sp.Matrix.eye(3)),  # Placeholder
        ('SLERP[q1_, q2_, t_]', lambda q1, q2, t: [(1-t)*q1i + t*q2i for q1i, q2i in zip(q1, q2)]),  # Simplified
        ('VelocityTransformation[omega_, v_, T_]', lambda omega, v, T: (omega, v)),
        ('ScewSymmetric[omega_]', lambda omega: sp.Matrix([[0, -omega[2], omega[1]], [omega[2], 0, -omega[0]], [-omega[1], omega[0], 0]])),
        ('TwistVector[v_, omega_]', lambda v, omega: sp.Matrix([v, omega])),
        ('WrenchVector[f_, tau_]', lambda f, tau: sp.Matrix([f, tau])),
        ('ManipulabilityIndex[J_]', lambda J: sqrt(sp.det(J*J.T))),
        ('ConditionNumber[J_]', lambda J: sp.Symbol('sigma_max')/sp.Symbol('sigma_min')),
        ('Singularity[det_J_]', lambda detJ: detJ == 0),
        ('Workspace[reachable_points_]', lambda pts: pts),
        ('DexterityMeasure[J_]', lambda J: 1/sp.Symbol('condition_number')),
        ('IsotropyIndex[J_]', lambda J: sp.Symbol('sigma_min')/sp.Symbol('sigma_max')),
        ('JointLimits[q_, q_min_, q_max_]', lambda q, qmin, qmax: qmin <= q <= qmax),
        ('SelfCollisionDetection[link_positions_]', lambda pos: False),
        ('PathPlanning[start_, goal_, obstacles_]', lambda s, g, obs: [s, g]),
        ('RRTAlgorithm[start_, goal_]', lambda s, g: [s, g]),
        ('PRMAlgorithm[nodes_, edges_]', lambda n, e: (n, e)),
        ('AStarPathPlanning[start_, goal_, heuristic_]', lambda s, g, h: [s, g]),
        ('DijkstraPathPlanning[start_, goal_]', lambda s, g: [s, g]),
        ('PotentialFieldMethod[q_, q_goal_, obstacles_]', lambda q, qg, obs: qg - q),
        ('TrajectoryGeneration[waypoints_, time_]', lambda wp, t: wp),
        
        # ===== DYNAMICS (30 functions) =====
        ('LagrangianMechanics[T_, V_]', lambda T, V: T - V),
        ('EulerLagrangeEquation[L_, q_]', lambda L, q: sp.diff(sp.diff(L, sp.diff(q, sp.Symbol('t'))), sp.Symbol('t')) - sp.diff(L, q)),
        ('KineticEnergy Robot[m_, v_, I_, omega_]', lambda m, v, I, omega: sp.Rational(1,2)*m*v**2 + sp.Rational(1,2)*I*omega**2),
        ('PotentialEnergyRobot[m_, g_, h_]', lambda m, g, h: m*g*h),
        ('NewtonEulerEquation[m_, a_, F_]', lambda m, a, F: m*a - F),
        ('InertiaMatrix[M_]', lambda M: M),
        ('CoriolisMatrix[C_]', lambda C: C),
        ('GravityVector[g_]', lambda g: g),
        ('JointTorque[tau_]', lambda tau: tau),
        ('InverseDynamics[M_, q_ddot_, C_, q_dot_, g_]', lambda M, qddot, C, qdot, g: M*qddot + C*qdot + g),
        ('ForwardDynamics[M_, tau_, C_, q_dot_, g_]', lambda M, tau, C, qdot, g: M.inv()*(tau - C*qdot - g)),
        ('RecursiveNewtonEuler[links_]', lambda links: links),
        ('ArticulatedBodyAlgorithm[tree_]', lambda tree: tree),
        ('CompositeRigidBodyAlgorithm[M_]', lambda M: M),
        ('CentrifugalForce[m_, omega_, r_]', lambda m, omega, r: m*omega**2*r),
        ('CoriolisForceRobot[m_, v_, omega_]', lambda m, v, omega: 2*m*v*omega),
        ('FrictionTorque[b_, q_dot_]', lambda b, qdot: b*qdot),
        ('DampingCoefficient[b_]', lambda b: b),
        ('SpringTorque[k_, q_]', lambda k, q: k*q),
        ('ActuatorDynamics[tau_motor_, gear_ratio_]', lambda tm, gr: tm*gr),
        ('MotorTorqueConstant[Kt_]', lambda Kt: Kt),
        ('BackEMFConstant[Ke_]', lambda Ke: Ke),
        ('MotorVoltageEquation[V_, R_, I_, Ke_, omega_]', lambda V, R, I, Ke, omega: V - R*I - Ke*omega),
        ('MotorCurrentEquation[tau_, Kt_]', lambda tau, Kt: tau/Kt),
        ('GearRatioRobot[N_in_, N_out_]', lambda Nin, Nout: Nin/Nout),
        ('EfficiencyTransmission[eta_]', lambda eta: eta),
        ('LoadInertiaReflected[I_load_, gear_ratio_]', lambda Iload, gr: Iload/gr**2),
        ('EquivalentInertia[I_motor_, I_load_reflected_]', lambda Im, Ilr: Im + Ilr),
        ('ZeromomentPoint[F_, M_]', lambda F, M: M/F),
        ('DynamicBalance[ZMP_, support_polygon_]', lambda ZMP, poly: ZMP in poly),
        
        # ===== CONTROL SYSTEMS (50 functions) =====
        ('PIDControl[Kp_, Ki_, Kd_, e_, int_e_, de_dt_]', lambda Kp, Ki, Kd, e, inte, dedt: Kp*e + Ki*inte + Kd*dedt),
        ('ProportionalControl[Kp_, e_]', lambda Kp, e: Kp*e),
        ('IntegralControl[Ki_, int_e_]', lambda Ki, inte: Ki*inte),
        ('DerivativeControl[Kd_, de_dt_]', lambda Kd, dedt: Kd*dedt),
        ('ErrorSignal[r_, y_]', lambda r, y: r - y),
        ('TransferFunctionControl[num_, den_]', lambda num, den: num/den),
        ('LaplaceDomain[time_function_]', lambda f: sp.Symbol('F(s)')),
        ('ClosedLoopTransferFunction[G_, H_]', lambda G, H: G/(1 + G*H)),
        ('OpenLoopGain[G_, H_]', lambda G, H: G*H),
        ('StabilityMargin[phase_, gain_]', lambda PM, GM: (PM, GM)),
        ('PhaseMargin[system_]', lambda sys: 180 + sp.Symbol('phase_at_gain_crossover')),
        ('GainMargin[system_]', lambda sys: -sp.Symbol('gain_at_phase_crossover')),
        ('RouthCriterion[coeffs_]', lambda c: True),  # Simplified stability test
        ('NyquistCriterion[GH_]', lambda GH: True),
        ('BodePlot[G_, omega_]', lambda G, omega: (20*log(abs(G), 10), sp.arg(G)*180/pi)),
        ('StepResponse[G_]', lambda G: sp.Symbol('y(t)')),
        ('ImpulseResponse[G_]', lambda G: sp.Symbol('h(t)')),
        ('RampResponse[G_]', lambda G: sp.Symbol('r(t)')),
        ('RiseTime[response_]', lambda resp: sp.Symbol('t_r')),
        ('SettlingTime[response_, tolerance_]', lambda resp, tol: sp.Symbol('t_s')),
        ('Overshoot[peak_, steady_state_]', lambda peak, ss: (peak - ss)/ss*100),
        ('SteadyStateError[r_, y_inf_]', lambda r, yinf: r - yinf),
        ('DampingRatioControl[zeta_]', lambda zeta: zeta),
        ('NaturalFrequencyControl[omega_n_]', lambda omegan: omegan),
        ('PolesControl[characteristic_eq_]', lambda char: sp.solve(char, sp.Symbol('s'))),
        ('ZerosControl[numerator_]', lambda num: sp.solve(num, sp.Symbol('s'))),
        ('RootLocusPlot[K_, poles_]', lambda K, poles: poles),
        ('StateSpaceModel[A_, B_, C_, D_]', lambda A, B, C, D: (A, B, C, D)),
        ('StateEquation[A_, x_, B_, u_]', lambda A, x, B, u: A*x + B*u),
        ('OutputEquation[C_, x_, D_, u_]', lambda C, x, D, u: C*x + D*u),
        ('Controllability[A_, B_]', lambda A, B: sp.Matrix.hstack(*[B, A*B, A**2*B])),
        ('Observability[A_, C_]', lambda A, C: sp.Matrix.vstack(*[C, C*A, C*A**2])),
        ('StateF eedback[K_, x_]', lambda K, x: -K*x),
        ('ObserverDesign[A_, L_, C_]', lambda A, L, C: A - L*C),
        ('KalmanFilter[A_, B_, C_, Q_, R_]', lambda A, B, C, Q, R: sp.Symbol('x_hat')),
        ('LQRControl[A_, B_, Q_, R_]', lambda A, B, Q, R: sp.Symbol('K')),
        ('LQGControl[A_, B_, C_, Q_, R_]', lambda A, B, C, Q, R: sp.Symbol('K')),
        ('H InfinityControl[plant_, weights_]', lambda plant, w: sp.Symbol('K')),
        ('RobustStability[uncertainty_, controller_]', lambda unc, K: True),
        ('AdaptiveControl[plant_, reference_model_]', lambda plant, ref: sp.Symbol('theta_hat')),
        ('ModelReferenceAdaptive[e_, gamma_]', lambda e, gamma: gamma*e),
        ('SelfTuningRegulator[params_]', lambda params: params),
        ('SliddingModeControl[s_, eta_]', lambda s, eta: -eta*sp.sign(s)),
        ('BacksteppingControl[error_dynamics_]', lambda edyn: sp.Symbol('u')),
        ('FeedbackLinearization[f_, g_, u_]', lambda f, g, u: (sp.Symbol('v') - f)/g),
        ('OptimalControl[J_, constraints_]', lambda J, cons: sp.Symbol('u_opt')),
        ('PontryaginMinimumPrinciple[H_]', lambda H: sp.diff(H, sp.Symbol('u'))),
        ('BellmanEquation[V_]', lambda V: sp.Symbol('V*')),
        ('ModelPredictiveControl[model_, horizon_, constraints_]', lambda model, N, cons: sp.Symbol('u_mpc')),
        ('RecedingHorizon[N_]', lambda N: N),
        
        # ===== SENSORS & PERCEPTION (20 functions) =====
        ('EncoderResolution[pulses_per_rev_]', lambda ppr: 360/ppr),
        ('PositionFromEncoder[counts_, resolution_]', lambda cnt, res: cnt*res),
        ('VelocityFromEncoder[delta_counts_, delta_t_, resolution_]', lambda dc, dt, res: dc*res/dt),
        ('AccelerometerOutput[a_]', lambda a: a),
        ('GyroscopeOutput[omega_]', lambda omega: omega),
        ('MagnetometerOutput[B_]', lambda B: B),
        ('IMUFusion[accel_, gyro_, mag_]', lambda a, g, m: (a, g, m)),
        ('ComplementaryFilter[accel_angle_, gyro_rate_, alpha_, dt_]', lambda aa, gr, alpha, dt: alpha*(aa + gr*dt) + (1 - alpha)*aa),
        ('KalmanFilterIMU[x_, P_, z_, Q_, R_]', lambda x, P, z, Q, R: x),
        ('MadgwickFilter[q_, gyro_, accel_, mag_]', lambda q, g, a, m: q),
        ('LiDARMeasurement[distance_, angle_]', lambda d, theta: (d*cos(theta), d*sin(theta))),
        ('PointCloudProcessing[points_]', lambda pts: pts),
        ('RANSAC[data_, model_]', lambda data, model: model),
        ('ICPAlgorithm[source_, target_]', lambda src, tgt: sp.Symbol('T')),
        ('VisualOdometry[images_]', lambda imgs: sp.Symbol('pose')),
        ('SLAM[sensors_, map_]', lambda sens, map: (sp.Symbol('pose'), map)),
        ('ParticleFilter[particles_, weights_, measurement_]', lambda p, w, z: p),
        ('OccupancyGrid[sensor_data_]', lambda data: sp.Matrix([[0]])),
        ('VoxelGrid[point_cloud_, resolution_]', lambda pc, res: pc),
        ('FeatureDetection[image_]', lambda img: [(x, y)]),
        ('FeatureMatching[features1_, features2_]', lambda f1, f2: list(zip(f1, f2))),
        
        # ===== ACTUATION (15 functions) =====
        ('ServoControl[PWM_]', lambda PWM: PWM),
        ('StepperMotorSteps[angle_, step_angle_]', lambda ang, stepang: ang/stepang),
        ('DCMotorSpeed[V_, Ke_]', lambda V, Ke: V/Ke),
        ('BrushlessMotorControl[phase_currents_]', lambda i: i),
        ('PWMDutyCycle[on_time_, period_]', lambda ton, T: ton/T*100),
        ('HBridgeControl[direction_, PWM_]', lambda dir, pwm: (dir, pwm)),
        ('CurrentControl[i_desired_, i_actual_, R_]', lambda id, ia, R: (id - ia)*R),
        ('VelocityControl[omega_desired_, omega_actual_, Kp_]', lambda od, oa, Kp: Kp*(od - oa)),
        ('PositionControl[theta_desired_, theta_actual_, Kp_]', lambda td, ta, Kp: Kp*(td - ta)),
        ('TorqueControl[tau_desired_, Kt_]', lambda td, Kt: td/Kt),
        ('PneumaticActuator[pressure_, area_]', lambda P, A: P*A),
        ('HydraulicActuator[pressure_, area_]', lambda P, A: P*A),
        ('LinearActuator[force_, displacement_]', lambda F, d: F*d),
        ('GripperForce[actuator_force_, mechanical_advantage_]', lambda F, MA: F*MA),
        ('ComplianceControl[F_desired_, F_actual_, K_]', lambda Fd, Fa, K: (Fd - Fa)/K),
    ]


def register():
    """Register all extended23 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""
MikoshiLang Extended Functions - Set 22
Medical sciences, pharmacology, and biomechanics
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi

def get_rules():
    """Extended set 22: Medical and pharmacological sciences (150 functions)"""
    t, C, V = symbols('t C V')
    
    return [
        # ===== PHARMACOKINETICS (40 functions) =====
        ('OneCompartmentModel[C0_, k_, t_]', lambda C0, k, t: C0*exp(-k*t)),
        ('TwoCompartmentModel[A_, B_, alpha_, beta_, t_]', lambda A, B, alpha, beta, t: A*exp(-alpha*t) + B*exp(-beta*t)),
        ('ThreeCompartmentModel[C1_, C2_, C3_, k1_, k2_, k3_, t_]', lambda C1, C2, C3, k1, k2, k3, t: C1*exp(-k1*t) + C2*exp(-k2*t) + C3*exp(-k3*t)),
        ('AbsorptionRateConstant[ka_]', lambda ka: ka),
        ('EliminationRateConstant[ke_]', lambda ke: ke),
        ('DistributionRateConstant[k12_]', lambda k12: k12),
        ('CentralCompartment[V1_, C1_]', lambda V1, C1: V1*C1),
        ('PeripheralCompartment[V2_, C2_]', lambda V2, C2: V2*C2),
        ('MassBalance[amount_in_, amount_out_]', lambda ain, aout: ain - aout),
        ('Bioavailability[AUC_po_, AUC_iv_, dose_iv_, dose_po_]', lambda AUCpo, AUCiv, Div, Dpo: (AUCpo/AUCiv)*(Div/Dpo)),
        ('FirstPassMetabolism[F_]', lambda F: 1 - F),
        ('ExtractionRatio[Cl_organ_, Q_]', lambda Cl, Q: Cl/Q),
        ('HepaticClearance[Q_, E_]', lambda Q, E: Q*E),
        ('RenalClearance[U_, P_]', lambda U, P: U/P),  # Urine conc / plasma conc * urine flow
        ('TotalBodyClearance[Cl_hepatic_, Cl_renal_, Cl_other_]', lambda Clh, Clr, Clo: Clh + Clr + Clo),
        ('IntrinsicClearance[Vmax_, Km_]', lambda Vmax, Km: Vmax/Km),
        ('UnboundFraction[fu_]', lambda fu: fu),
        ('ProteinBinding[fu_]', lambda fu: 1 - fu),
        ('VolumeOfDistributionSteadyState[Vd_]', lambda Vd: Vd),
        ('ApparentVolumeOfDistribution[amount_, concentration_]', lambda amt, C: amt/C),
        ('MeanResidenceTime[AUMC_, AUC_]', lambda AUMC, AUC: AUMC/AUC),
        ('MeanAbsorptionTime[MAT_]', lambda MAT: MAT),
        ('TimeToMaxConcentration[ka_, ke_]', lambda ka, ke: log(ka/ke)/(ka - ke)),
        ('MaxConcentration[dose_, F_, V_, ka_, ke_]', lambda D, F, V, ka, ke: D*F/V*ka/(ka - ke)),
        ('AUCInfinity[C0_, ke_]', lambda C0, ke: C0/ke),
        ('AUCTrapezoid[C1_, C2_, delta_t_]', lambda C1, C2, dt: (C1 + C2)/2*dt),
        ('SteadyStateConcentrationAverage[dose_, tau_, Cl_]', lambda D, tau, Cl: D/(tau*Cl)),
        ('SteadyStateConcentrationMax[Css_avg_, ka_, ke_, tau_]', lambda Cssavg, ka, ke, tau: Cssavg*ka/(ka - ke)*(1 - exp(-ke*tau))),
        ('SteadyStateConcentrationMin[Css_max_, ke_, tau_]', lambda Cssmax, ke, tau: Cssmax*exp(-ke*tau)),
        ('FluctuationDegree[Css_max_, Css_min_]', lambda Cmax, Cmin: (Cmax - Cmin)/Cmin*100),
        ('AccumulationIndex[Css_, C1_]', lambda Css, C1: Css/C1),
        ('LoadingDosePK[Vd_, Css_target_]', lambda Vd, Csst: Vd*Csst),
        ('MaintenanceDosePK[Cl_, Css_, tau_]', lambda Cl, Css, tau: Cl*Css*tau),
        ('DoseAdjustmentRenal[dose_normal_, CLcr_, CLcr_normal_]', lambda Dn, CLcr, CLcrn: Dn*CLcr/CLcrn),
        ('CrockcroftGault[age_, weight_, Scr_, sex_]', lambda age, wt, Scr, sex: (140 - age)*wt/(72*Scr)*(0.85 if str(sex) == 'F' else 1)),
        ('MDRDFormula[Scr_, age_, sex_]', lambda Scr, age, sex: 175*Scr**(-1.154)*age**(-0.203)*(0.742 if str(sex) == 'F' else 1)),
        ('ChildPughScore[bili_, alb_, PT_, ascites_, enceph_]', lambda bili, alb, PT, asc, enc: bili + alb + PT + asc + enc),
        ('TherapeuticRange[Cmin_, Cmax_]', lambda Cmin, Cmax: (Cmin, Cmax)),
        ('ToxicConcentration[Ctox_]', lambda Ctox: Ctox),
        ('SubtherapeuticConcentration[Csub_]', lambda Csub: Csub),
        
        # ===== PHARMACODYNAMICS (30 functions) =====
        ('DoseResponseCurve[dose_, response_]', lambda D, R: (D, R)),
        ('EC50[dose_]', lambda D: D),  # Effective concentration for 50% response
        ('ED50[dose_]', lambda D: D),  # Effective dose for 50% response
        ('IC50[concentration_]', lambda C: C),  # Inhibitory concentration 50%
        ('LD50[dose_]', lambda D: D),  # Lethal dose 50%
        ('TherapeuticWindowPD[ED50_, TD50_]', lambda ED50, TD50: TD50/ED50),
        ('EMaxModel[E0_, Emax_, C_, EC50_]', lambda E0, Emax, C, EC50: E0 + Emax*C/(EC50 + C)),
        ('SigmoidEMax[E0_, Emax_, C_, EC50_, gamma_]', lambda E0, Emax, C, EC50, gamma: E0 + Emax*C**gamma/(EC50**gamma + C**gamma)),
        ('LinearModel[E0_, slope_, C_]', lambda E0, m, C: E0 + m*C),
        ('LogLinearModel[E0_, slope_, log_C_]', lambda E0, m, logC: E0 + m*logC),
        ('InhibitoryEMax[E0_, Imax_, C_, IC50_]', lambda E0, Imax, C, IC50: E0 - Imax*C/(IC50 + C)),
        ('AgonistActivity[efficacy_, EC50_]', lambda eff, EC50: eff/EC50),
        ('AntagonismCompetitive[Kd_agonist_, Kantagonist_, antagonist_conc_]', lambda Kda, Ka, Ca: Kda*(1 + Ca/Ka)),
        ('AntagonismNoncompetitive[Emax_, antagonist_]', lambda Emax, antag: Emax/(1 + antag)),
        ('PartialAgonist[E_max_partial_, E_max_full_]', lambda Emaxp, Emaxf: Emaxp/Emaxf),
        ('InverseAgonist[basal_, inhibition_]', lambda basal, inhib: basal - inhib),
        ('Desensitization[response_initial_, response_sustained_]', lambda Ri, Rs: (Ri - Rs)/Ri),
        ('Tachyphylaxis[response_decrease_]', lambda dec: dec),
        ('Tolerance[dose_initial_, dose_required_]', lambda Di, Dr: Dr/Di),
        ('Sensitization[response_increase_]', lambda inc: inc),
        ('Hysteresis[C_, E_]', lambda C, E: E - sp.Symbol('E_predicted')),
        ('EffectCompartmentModel[Ce_, keo_]', lambda Ce, keo: sp.diff(Ce, sp.Symbol('t')) + keo*Ce),
        ('ReceptorOccupancyTheory[Kd_, L_]', lambda Kd, L: L/(Kd + L)),
        ('SparereceptorTheory[occupancy_, response_]', lambda occ, resp: resp > occ),
        ('AllosericModulation[binding_, modulator_]', lambda bind, mod: bind*mod),
        ('Synergism[effect_combined_, effect_individual_]', lambda Ec, Ei: Ec > sum(Ei)),
        ('Antagonism[effect_combined_, effect_individual_]', lambda Ec, Ei: Ec < sum(Ei)),
        ('AdditiveEffect[E1_, E2_]', lambda E1, E2: E1 + E2),
        ('PotentiationEffect[E_enhanced_]', lambda Ee: Ee),
        ('FunctionalAntagonism[effect1_, effect2_]', lambda E1, E2: E1 - E2),
        
        # ===== TOXICOLOGY (25 functions) =====
        ('LethalDose[LD50_]', lambda LD50: LD50),
        ('NoObservedAdverseEffectLevel[NOAEL_]', lambda NOAEL: NOAEL),
        ('LowestObservedAdverseEffectLevel[LOAEL_]', lambda LOAEL: LOAEL),
        ('AcceptableDailyIntake[NOAEL_, safety_factor_]', lambda NOAEL, SF: NOAEL/SF),
        ('ReferenceOose[RfD_]', lambda RfD: RfD),
        ('MarginOfSafety[NOAEL_, exposure_]', lambda NOAEL, exp: NOAEL/exp),
        ('HazardQuotient[exposure_, RfD_]', lambda exp, RfD: exp/RfD),
        ('CumulativeRisk[HQ_]', lambda HQ: sum(HQ)),
        ('BenchmarkDose[BMD_]', lambda BMD: BMD),
        ('PointOfDeparture[POD_]', lambda POD: POD),
        ('DoseAdditionModel[doses_]', lambda doses: sum(doses)),
        ('ResponseAdditionModel[responses_]', lambda resp: 1 - sp.Mul(*[1 - ri for ri in resp])),
        ('InteractionIndex[toxicity_combined_, toxicity_expected_]', lambda Tc, Te: Tc/Te),
        ('CytotoxicityAssay[viable_cells_, total_cells_]', lambda viable, total: (1 - viable/total)*100),
        ('GenotoxicityIndex[damage_]', lambda dam: dam),
        ('MutagenicityTest[mutations_]', lambda mut: mut),
        ('CarcinogenicityRisk[slope_factor_, dose_]', lambda SF, D: SF*D),
        ('ExposureLimit[TLV_]', lambda TLV: TLV),  # Threshold limit value
        ('PermissibleExposureLimit[PEL_]', lambda PEL: PEL),
        ('ImmediateDangerous ToLifeOrHealth[IDLH_]', lambda IDLH: IDLH),
        ('AcuteExposureGuideline[AEGL_]', lambda AEGL: AEGL),
        ('EmergencyResponsePlanningGuideline[ERPG_]', lambda ERPG: ERPG),
        ('BiomonitoringEquivalent[BE_]', lambda BE: BE),
        ('InternalDose[absorbed_, metabolized_]', lambda abs, met: abs - met),
        ('TargetOrganDose[internal_dose_, distribution_]', lambda ID, dist: ID*dist),
        
        # ===== CLINICAL MEASUREMENTS (30 functions) =====
        ('BMI[weight_, height_]', lambda wt, ht: wt/ht**2),  # Alias for BodyMassIndex
        ('BodyMassIndex[weight_, height_]', lambda wt, ht: wt/ht**2),
        ('BodySurfaceArea[height_, weight_]', lambda ht, wt: sqrt(ht*wt/3600)),  # Mosteller
        ('DuBoisFormula[height_, weight_]', lambda ht, wt: 0.007184*wt**0.425*ht**0.725),
        ('IdealBodyWeight[height_, sex_]', lambda ht, sex: 50 + 2.3*(ht - 60) if str(sex) == 'M' else 45.5 + 2.3*(ht - 60)),
        ('LeanBodyMass[weight_, height_, sex_]', lambda wt, ht, sex: 0.407*wt + 0.267*ht - 19.2 if str(sex) == 'M' else 0.252*wt + 0.473*ht - 48.3),
        ('BasalMetabolicRate[weight_, height_, age_, sex_]', lambda wt, ht, age, sex: 10*wt + 6.25*ht - 5*age + 5 if str(sex) == 'M' else 10*wt + 6.25*ht - 5*age - 161),
        ('TotalEnergyExpenditure[BMR_, PAL_]', lambda BMR, PAL: BMR*PAL),  # Physical activity level
        ('CardiacOutput[HR_, SV_]', lambda HR, SV: HR*SV/1000),  # L/min
        ('StrokeVolume[EDV_, ESV_]', lambda EDV, ESV: EDV - ESV),
        ('EjectionFraction[SV_, EDV_]', lambda SV, EDV: SV/EDV*100),
        ('MeanArterialPressure[SBP_, DBP_]', lambda SBP, DBP: DBP + (SBP - DBP)/3),
        ('PulsePressure[SBP_, DBP_]', lambda SBP, DBP: SBP - DBP),
        ('SystemicVascularResistance[MAP_, CVP_, CO_]', lambda MAP, CVP, CO: (MAP - CVP)/CO*80),  # dynes·s/cm⁵
        ('PulmonaryVascularResistance[MPAP_, PCWP_, CO_]', lambda MPAP, PCWP, CO: (MPAP - PCWP)/CO*80),
        ('RespiratoryRate[breaths_per_min_]', lambda RR: RR),
        ('TidalVolume[TV_]', lambda TV: TV),  # mL
        ('MinuteVentilation[TV_, RR_]', lambda TV, RR: TV*RR/1000),  # L/min
        ('AlveolarVentilation[TV_, dead_space_, RR_]', lambda TV, DS, RR: (TV - DS)*RR/1000),
        ('OxygenSaturation[SpO2_]', lambda SpO2: SpO2),
        ('PaO2FiO2Ratio[PaO2_, FiO2_]', lambda PaO2, FiO2: PaO2/FiO2),
        ('AlveolarArterialGradient[PAO2_, PaO2_]', lambda PAO2, PaO2: PAO2 - PaO2),
        ('AlveolarGasEquation[FiO2_, PaCO2_]', lambda FiO2, PaCO2: (FiO2*713) - PaCO2/0.8),
        ('RespiratoryQuotient[VCO2_, VO2_]', lambda VCO2, VO2: VCO2/VO2),
        ('OxygenConsumption[VO2_]', lambda VO2: VO2),  # mL/min
        ('CarbonDioxideProduction[VCO2_]', lambda VCO2: VCO2),
        
        # ===== BIOMECHANICS (25 functions) =====
        ('ForceBodyWeight[mass_, g_]', lambda m, g: m*g),
        ('TorqueJoint[F_, r_]', lambda F, r: F*r),
        ('MomentArm[perpendicular_distance_]', lambda d: d),
        ('MechanicalAdvantageBody[F_out_, F_in_]', lambda Fout, Fin: Fout/Fin),
        ('GroundReactionForce[mass_, acceleration_]', lambda m, a: m*(sp.Symbol('g') + a)),
        ('ImpulseCollision[F_, delta_t_]', lambda F, dt: F*dt),
        ('WorkMechanical[F_, d_]', lambda F, d: F*d),
        ('PowerOutput[work_, time_]', lambda W, t: W/t),
        ('StrainTissue[delta_L_, L0_]', lambda dL, L0: dL/L0),
        ('StressTissue[F_, A_]', lambda F, A: F/A),
        ('YoungsModulusTissue[stress_, strain_]', lambda sig, eps: sig/eps),
        ('ViscoelasticResponse[stress_, time_]', lambda sig, t: sig*exp(-t/sp.Symbol('tau'))),
        ('StressRelaxation[sigma0_, t_, tau_]', lambda sig0, t, tau: sig0*exp(-t/tau)),
        ('CreepCompliance[strain_, time_]', lambda eps, t: eps/sp.Symbol('sigma0')*(1 - exp(-t/sp.Symbol('tau')))),
        ('BoneStrength[density_, geometry_]', lambda rho, geom: rho*geom),
        ('FractureLoad[stress_ultimate_, area_]', lambda sigult, A: sigult*A),
        ('GaitCycle[stance_, swing_]', lambda stance, swing: stance + swing),
        ('CadenceWalking[steps_per_min_]', lambda cadence: cadence),
        ('StrideLength[distance_, steps_]', lambda d, steps: d/steps),
        ('DoubleSupport[time_both_feet_]', lambda t: t),
        ('CenterOfMass[masses_, positions_]', lambda m, x: sum(mi*xi for mi, xi in zip(m, x))/sum(m)),
        ('BalanceIndex[sway_]', lambda sway: sway),
        ('PosturalStability[COP_displacement_]', lambda COP: COP),
        ('JumpHeight[flight_time_]', lambda t: sp.Symbol('g')*t**2/8),
        ('PowerClean[work_, time_]', lambda W, t: W/t),
    ]


def register():
    """Register all extended22 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""
MikoshiLang Extended Functions - Set 33
Agriculture, food science, sports science, and miscellaneous engineering
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi, sin, cos

def get_rules():
    """Extended set 33: Agriculture, food, sports, and specialized engineering (100 functions)"""
    x, y, t = symbols('x y t')
    
    return [
        # ===== AGRICULTURE & SOIL SCIENCE (25 functions) =====
        ('SoilMoisture[theta_]', lambda theta: theta),
        ('FieldCapacity[FC_]', lambda FC: FC),
        ('WiltingPoint[WP_]', lambda WP: WP),
        ('AvailableWaterCapacity[FC_, WP_]', lambda FC, WP: FC - WP),
        ('SoilWaterDeficit[FC_, current_]', lambda FC, curr: FC - curr),
        ('IrrigationRequirement[ET_, P_, Delta_S_]', lambda ET, P, dS: ET - P + dS),
        ('EvapotranspirationPenman[Rn_, T_, u_, e_s_, e_a_]', lambda Rn, T, u, es, ea: sp.Symbol('ET')),
        ('ReferenceET[T_, RH_, wind_, radiation_]', lambda T, RH, w, R: sp.Symbol('ET0')),
        ('CropCoefficient[Kc_]', lambda Kc: Kc),
        ('CropEvapotranspiration[ET0_, Kc_]', lambda ET0, Kc: ET0*Kc),
        ('GrowingDegreeDays[T_max_, T_min_, T_base_]', lambda Tmax, Tmin, Tbase: (Tmax + Tmin)/2 - Tbase),
        ('ChillingHours[T_]', lambda T: 1 if T < 7 else 0),
        ('HeatUnits[T_, T_base_]', lambda T, Tbase: max(0, T - Tbase)),
        ('LeafAreaIndex[leaf_area_, ground_area_]', lambda La, Ga: La/Ga),
        ('PhotosyntheticRate[light_, CO2_, T_]', lambda L, CO2, T: sp.Symbol('P')),
        ('SoilpH[H_concentration_]', lambda H: -log(H, 10)),
        ('CationExchangeCapacity[CEC_]', lambda CEC: CEC),
        ('SoilOrganicMatter[C_content_]', lambda C: C*1.724),  # Van Bemmelen factor
        ('BulkDensity[mass_, volume_]', lambda m, V: m/V),
        ('Porosity[bulk_density_, particle_density_]', lambda rhob, rhop: (1 - rhob/rhop)*100),
        ('HydraulicConductivity[K_]', lambda K: K),
        ('InfiltrationRate[I_]', lambda I: I),
        ('RunoffCoefficient[C_]', lambda C: C),
        ('SoilErosionUSLE[R_, K_, LS_, C_, P_]', lambda R, K, LS, C, P: R*K*LS*C*P),
        ('NutrientUptakeRate[C_, K_m_, V_max_]', lambda C, Km, Vmax: Vmax*C/(Km + C)),
        
        # ===== FOOD SCIENCE (25 functions) =====
        ('WaterActivity[P_, P0_]', lambda P, P0: P/P0),
        ('MoistureContent[wet_basis_]', lambda wb: wb/(1 - wb)*100),
        ('SolidContent[moisture_]', lambda m: 100 - m),
        ('Brix[sucrose_percent_]', lambda s: s),
        ('AcidityTA[acid_meq_, volume_]', lambda meq, V: meq/V),
        ('pHFood[H_]', lambda H: -log(H, 10)),
        ('TitrableAcidity[NaOH_vol_, normality_, sample_]', lambda V, N, s: V*N/s*100),
        ('BrowningIndex[A_420_]', lambda A420: A420*100),
        ('ColorDifference[L1_, a1_, b1_, L2_, a2_, b2_]', lambda L1, a1, b1, L2, a2, b2: sqrt((L2-L1)**2 + (a2-a1)**2 + (b2-b1)**2)),
        ('TextureHardness[force_]', lambda F: F),
        ('Viscosity[shear_stress_, shear_rate_]', lambda tau, gamma: tau/gamma),
        ('PowerLawFluid[K_, n_, gamma_]', lambda K, n, gamma: K*gamma**n),
        ('YieldStress[tau_0_]', lambda tau0: tau0),
        ('RheologyIndex[n_]', lambda n: n),
        ('FreezingPointDepression[molality_]', lambda m: 1.86*m),
        ('DValue[T_]', lambda T: sp.Symbol('D')),  # Decimal reduction time
        ('ZValue[microorganism_]', lambda m: sp.Symbol('z')),
        ('FValue[D_, T_ref_, z_]', lambda D, Tref, z: D*10**((sp.Symbol('T') - Tref)/z)),
        ('ThermalDeathTime[F0_]', lambda F0: F0),
        ('PasteurizationUnit[PU_]', lambda PU: PU),
        ('ShelfLifePrediction[k_, T_]', lambda k, T: log(2)/k),
        ('ArrheniusShelfLife[k0_, Ea_, R_, T_]', lambda k0, Ea, R, T: k0*exp(-Ea/(R*T))),
        ('OxygenPermeability[P_, A_, delta_x_, delta_p_]', lambda P, A, dx, dp: P*A*dp/dx),
        ('WaterVaporTransmissionRate[WVTR_]', lambda WVTR: WVTR),
        ('PackagingBarrierFactor[actual_rate_, ideal_rate_]', lambda actual, ideal: ideal/actual),
        
        # ===== SPORTS SCIENCE (20 functions) =====
        ('VO2Max[ml_min_, kg_]', lambda V, kg: V/kg),
        ('MaxHeartRate[age_]', lambda age: 220 - age),
        ('TargetHeartRateZone[max_HR_, intensity_]', lambda maxHR, i: maxHR*i),
        ('KarvonenFormula[max_HR_, resting_HR_, intensity_]', lambda maxHR, restHR, i: (maxHR - restHR)*i + restHR),
        ('CaloriesBurned[MET_, weight_, duration_]', lambda MET, w, d: MET*w*d/60),
        ('RunningPower[mass_, velocity_, grade_, Cair_]', lambda m, v, gr, Cair: m*9.81*v*gr + Cair*v**3),
        ('CriticalPower[W_max_, AWC_, t_]', lambda Wmax, AWC, t: Wmax - AWC/t),
        ('AnaerobicWorkCapacity[AWC_]', lambda AWC: AWC),
        ('FunctionalThresholdPower[FTP_]', lambda FTP: FTP),
        ('PowerToWeightRatio[watts_, kg_]', lambda W, kg: W/kg),
        ('CyclingEfficiency[power_out_, power_in_]', lambda Pout, Pin: Pout/Pin*100),
        ('SwimmingPace[time_, distance_]', lambda t, d: t/d),
        ('RunningEconomy[VO2_, velocity_]', lambda VO2, v: VO2/v),
        ('StrideLength[distance_, steps_]', lambda d, s: d/s),
        ('StrideFrequency[steps_, time_]', lambda s, t: s/t),
        ('GroundContactTime[t_contact_]', lambda tc: tc),
        ('VerticalOscillation[height_]', lambda h: h),
        ('LactateThreshold[lactate_]', lambda lac: lac),
        ('RespiratoryExchangeRatio[VCO2_, VO2_]', lambda VCO2, VO2: VCO2/VO2),
        ('TrainingLoad[duration_, intensity_]', lambda d, i: d*i),
        
        # ===== TEXTILE ENGINEERING (15 functions) =====
        ('YarnCount[weight_, length_]', lambda w, l: l/w),
        ('Tex[grams_, km_]', lambda g, km: g/km),
        ('Denier[grams_, 9000m_]', lambda g, L: g/L*9000),
        ('FabricGSM[weight_, area_]', lambda w, A: w/A*1000),  # Grams per square meter
        ('ThreadsPerInch[TPI_]', lambda TPI: TPI),
        ('FabricCoverFactor[threads_, yarn_diameter_]', lambda n, d: n*d),
        ('TensilleStrength[force_, area_]', lambda F, A: F/A),
        ('Elongation[delta_L_, L0_]', lambda dL, L0: dL/L0*100),
        ('YoungModulusTextile[stress_, strain_]', lambda sig, eps: sig/eps),
        ('FabricThickness[t_]', lambda t: t),
        ('AirPermeability[Q_, A_, delta_P_]', lambda Q, A, dP: Q/(A*dP)),
        ('MoistureRegain[W_water_, W_dry_]', lambda Ww, Wd: (Ww - Wd)/Wd*100),
        ('WaterAbsorption[W_wet_, W_dry_]', lambda Wwet, Wdry: (Wwet - Wdry)/Wdry*100),
        ('ThermalResistance[thickness_, conductivity_]', lambda t, k: t/k),
        ('ClothingInsulation[Clo_]', lambda Clo: Clo*0.155),  # Clo to m²·K/W
        
        # ===== PETROLEUM ENGINEERING (15 functions) =====
        ('OilGravityAPI[SG_]', lambda SG: 141.5/SG - 131.5),
        ('GasOilRatio[gas_volume_, oil_volume_]', lambda Vg, Vo: Vg/Vo),
        ('FormationVolumeFactor[V_reservoir_, V_surface_]', lambda Vr, Vs: Vr/Vs),
        ('ReservoirPressure[depth_, gradient_]', lambda d, g: d*g),
        ('DarcyLaw[Q_, k_, A_, mu_, L_, delta_P_]', lambda Q, k, A, mu, L, dP: Q - k*A*dP/(mu*L)),
        ('Permeability[Q_, mu_, L_, A_, delta_P_]', lambda Q, mu, L, A, dP: Q*mu*L/(A*dP)),
        ('PorousMediaFlow[k_, phi_]', lambda k, phi: k*phi),
        ('RecoveryFactor[produced_, OOIP_]', lambda prod, OOIP: prod/OOIP*100),
        ('OilInPlace[A_, h_, phi_, Sw_, Bo_]', lambda A, h, phi, Sw, Bo: 7758*A*h*phi*(1 - Sw)/Bo),
        ('GasInPlace[A_, h_, phi_, Sw_, Bg_]', lambda A, h, phi, Sw, Bg: 43560*A*h*phi*(1 - Sw)/Bg),
        ('WaterSaturation[Sw_]', lambda Sw: Sw),
        ('ResidualOilSaturation[Sor_]', lambda Sor: Sor),
        ('CapillaryPressure[sigma_, theta_, r_]', lambda sig, theta, r: 2*sig*cos(theta)/r),
        ('RelativePermeability[k_rel_]', lambda krel: krel),
        ('WellProductivityIndex[Q_, P_res_, P_well_]', lambda Q, Pres, Pwell: Q/(Pres - Pwell)),
    ]


def register():
    """Register all extended33 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""Numerical evaluation and constants."""

from __future__ import annotations
from typing import Any
import sympy as sp
from .expr import Expr, Symbol
from .math_ops import to_sympy, from_sympy

# Mathematical constants
Pi = Symbol("Pi")
E = Symbol("E")
I = Symbol("I")
Infinity = Symbol("Infinity")
GoldenRatio = Symbol("GoldenRatio")


def N(expr, precision: int = 15) -> Any:
    """Evaluate expression numerically to given precision."""
    return numerical(expr, precision)


def numerical(expr, precision: int = 15) -> Any:
    """Evaluate expression to floating point with given precision."""
    if isinstance(expr, (int, float)):
        return float(expr)
    if isinstance(expr, Symbol):
        _constants = {
            "Pi": lambda p: float(sp.pi.evalf(p)),
            "E": lambda p: float(sp.E.evalf(p)),
            "I": lambda p: complex(0, 1),
            "Infinity": lambda p: float("inf"),
            "GoldenRatio": lambda p: float(sp.GoldenRatio.evalf(p)),
        }
        if expr.name in _constants:
            return _constants[expr.name](precision)
        return expr
    try:
        sexpr = to_sympy(expr)
        result = sexpr.evalf(precision)
        if result.is_real:
            return float(result)
        if result.is_number:
            return complex(result)
        return from_sympy(result)
    except Exception:
        return expr



"""Tests for chemistry module."""

import pytest
from mikoshilang.chemistry import Element, AtomicMass, ElectronConfiguration, MolecularMass, BalanceEquation
from mikoshilang.expr import Expr


class TestElements:
    def test_hydrogen(self):
        e = Element("H")
        assert isinstance(e, Expr)
        # Check it has the right data
        assert e.head == "List"

    def test_oxygen(self):
        mass = AtomicMass("O")
        assert abs(mass - 15.999) < 0.01

    def test_iron_config(self):
        config = ElectronConfiguration("Fe")
        assert config == "[Ar] 3d6 4s2"

    def test_by_number(self):
        e = Element(1)
        assert e.head == "List"

    def test_by_name(self):
        mass = AtomicMass("Carbon")
        assert abs(mass - 12.011) < 0.01

    def test_all_118_elements(self):
        for i in range(1, 119):
            e = Element(i)
            assert e.head == "List"

    def test_unknown_element(self):
        with pytest.raises(ValueError):
            Element("Xx")

    def test_gold(self):
        assert abs(AtomicMass("Au") - 196.97) < 0.01

    def test_uranium(self):
        assert abs(AtomicMass("U") - 238.03) < 0.01

    def test_helium(self):
        config = ElectronConfiguration("He")
        assert config == "1s2"


class TestMolecularMass:
    def test_water(self):
        mass = MolecularMass("H2O")
        assert abs(mass - 18.015) < 0.01

    def test_glucose(self):
        mass = MolecularMass("C6H12O6")
        assert abs(mass - 180.156) < 0.1

    def test_co2(self):
        mass = MolecularMass("CO2")
        assert abs(mass - 44.009) < 0.01

    def test_nacl(self):
        mass = MolecularMass("NaCl")
        expected = AtomicMass("Na") + AtomicMass("Cl")
        assert abs(mass - expected) < 0.01

    def test_single_element(self):
        mass = MolecularMass("He")
        assert abs(mass - 4.003) < 0.01


class TestBalanceEquation:
    def test_h2_o2(self):
        result = BalanceEquation("H2 + O2 -> H2O")
        assert "2H2" in result
        assert "2H2O" in result

    def test_already_balanced(self):
        # Simple case
        result = BalanceEquation("H2 + Cl2 -> HCl")
        assert "2HCl" in result

    def test_combustion(self):
        result = BalanceEquation("CH4 + O2 -> CO2 + H2O")
        # Should balance to CH4 + 2O2 -> CO2 + 2H2O
        assert "2" in result



"""
MikoshiLang Extended Functions - Set 30
Acoustics, audio engineering, psychoacoustics, and vibration analysis
"""
import sympy as sp
from sympy import symbols, sqrt, sin, cos, exp, log, pi

def get_rules():
    """Extended set 30: Acoustics and audio (150 functions)"""
    f, t, A = symbols('f t A')
    
    return [
        # ===== ACOUSTICS FUNDAMENTALS (35 functions) =====
        ('SoundPressureLevel[p_, p_ref_]', lambda p, pref: 20*log(p/pref, 10)),
        ('SoundIntensityLevel[I_, I_ref_]', lambda I, Iref: 10*log(I/Iref, 10)),
        ('SoundPowerLevel[W_, W_ref_]', lambda W, Wref: 10*log(W/Wref, 10)),
        ('AcousticIntensity[p_, rho_, c_]', lambda p, rho, c: p**2/(rho*c)),
        ('AcousticImpedanceWave[rho_, c_]', lambda rho, c: rho*c),
        ('SpeedOfSoundAir[T_]', lambda T: 331.3*sqrt(1 + T/273.15)),
        ('SpeedOfSoundWater[T_, S_, d_]', lambda T, S, d: 1449 + 4.6*T - 0.055*T**2 + 0.00029*T**3 + (1.34 - 0.01*T)*(S - 35) + 0.016*d),
        ('WavelengthAcoustic[c_, f_]', lambda c, f: c/f),
        ('FrequencyAcoustic[c_, lambda_]', lambda c, lam: c/lam),
        ('ParticleVelocity[p_, rho_, c_]', lambda p, rho, c: p/(rho*c)),
        ('ParticleDisplacement[v_, omega_]', lambda v, omega: v/omega),
        ('AcousticEnergy Density[p_, rho_, c_]', lambda p, rho, c: p**2/(rho*c**2)),
        ('RayleighDistance[D_, lambda_]', lambda D, lam: D**2/(4*lam)),
        ('NearFieldDistance[D_, lambda_]', lambda D, lam: D**2/(4*lam)),
        ('FarFieldApproximation[r_, D_, lambda_]', lambda r, D, lam: r > D**2/(4*lam)),
        ('FrequencyRatio[f1_, f2_]', lambda f1, f2: log(f1/f2, 2)),  # Octaves
        ('MusicalInterval[f1_, f2_]', lambda f1, f2: 1200*log(f1/f2, 2)),  # Cents
        ('HarmonicSeries[f0_, n_]', lambda f0, n: n*f0),
        ('Overtone[f0_, n_]', lambda f0, n: (n + 1)*f0),
        ('FundamentalFrequency[harmonics_]', lambda h: sp.gcd(*h) if len(h) > 1 else h[0]),
        ('Inharmonicity[f_partial_, n_, f0_]', lambda fp, n, f0: (fp - n*f0)/(n*f0)),
        ('BeatFrequency[f1_, f2_]', lambda f1, f2: abs(f1 - f2)),
        ('CombinationTone[f1_, f2_]', lambda f1, f2: abs(f1 - f2)),  # Difference tone
        ('SummationTone[f1_, f2_]', lambda f1, f2: f1 + f2),
        ('DopplerShift[f0_, v_source_, v_observer_, c_]', lambda f0, vs, vo, c: f0*(c + vo)/(c - vs)),
        ('SonicBoom[M_]', lambda M: M > 1),
        ('MachWaveAngle[M_]', lambda M: sp.asin(1/M)),
        ('StandingWave[A_, k_, x_, omega_, t_]', lambda A, k, x, omega, t: 2*A*sin(k*x)*cos(omega*t)),
        ('NodePosition[n_, lambda_]', lambda n, lam: n*lam/2),
        ('AntinodePosition[n_, lambda_]', lambda n, lam: (2*n + 1)*lam/4),
        ('ResonanceFrequency[L_, n_, c_]', lambda L, n, c: n*c/(2*L)),
        ('FundamentalMode[L_, c_]', lambda L, c: c/(2*L)),
        ('OpenPipeResonance[L_, n_, c_]', lambda L, n, c: n*c/(2*L)),
        ('ClosedPipeResonance[L_, n_, c_]', lambda L, n, c: (2*n - 1)*c/(4*L)),
        ('HelmholtzResonator[V_, A_, L_]', lambda V, A, L: sp.Symbol('c')/( 2*pi)*sqrt(A/(V*L))),
        
        # ===== ROOM ACOUSTICS (30 functions) =====
        ('SabineReverbTime[V_, A_]', lambda V, A: 0.161*V/A),
        ('EyringReverbTime[V_, S_, alpha_]', lambda V, S, alpha: -0.161*V/(S*log(1 - alpha))),
        ('MilngtonSetteFormula[V_, S_i_, alpha_i_]', lambda V, Si, alphai: -0.161*V/sum(Sii*log(1 - alphai) for Sii, alphai in zip(Si, alphai))),
        ('AbsorptionCoefficient[alpha_]', lambda alpha: alpha),
        ('SabinAbsorption[S_, alpha_]', lambda S, alpha: S*alpha),
        ('TotalRoomAbsorption[materials_]', lambda m: sum(m)),
        ('RoomConstant[S_, alpha_]', lambda S, alpha: S*alpha/(1 - alpha)),
        ('CriticalDistance[R_, Q_]', lambda R, Q: sqrt(R*Q/(16*pi))),
        ('DirectToreverbRatio[r_, R_]', lambda r, R: 10*log(R/(16*pi*r**2) + 1, 10)),
        ('SoundDecayRate[T60_]', lambda T60: 60/T60),  # dB/s
        ('EarlyDecayTime[EDT_]', lambda EDT: EDT),
        ('Clarity[C80_]', lambda C80: C80),
        ('Definition[D50_]', lambda D50: D50),
        ('SpeechTransmissionIndex[STI_]', lambda STI: STI),
        ('ArticulationIndex[AI_]', lambda AI: AI),
        ('ModulationTransferFunction[m_]', lambda m: m),
        ('RapidSpeechTransmissionIndex[RASTI_]', lambda RASTI: RASTI),
        ('SoundStrengthG[I_, I_ref_]', lambda I, Iref: 10*log(I/Iref, 10)),
        ('LateralEnergyFraction[LF_]', lambda LF: LF),
        ('InterauralCrossCorrelation[IACC_]', lambda IACC: IACC),
        ('CenterTime[ts_]', lambda ts: ts),
        ('RoomGain[direct_, reverb_]', lambda d, r: 10*log((d + r)/d, 10)),
        ('NoiseReductionCoefficient[alpha_250_, alpha_500_, alpha_1000_, alpha_2000_]', lambda a250, a500, a1000, a2000: (a250 + a500 + a1000 + a2000)/4),
        ('SoundAbsorptionAverage[alphas_]', lambda alphas: sum(alphas)/len(alphas)),
        ('RoomModeFrequency[nx_, ny_, nz_, Lx_, Ly_, Lz_, c_]', lambda nx, ny, nz, Lx, Ly, Lz, c: c/2*sqrt((nx/Lx)**2 + (ny/Ly)**2 + (nz/Lz)**2)),
        ('AxialMode[n_, L_, c_]', lambda n, L, c: n*c/(2*L)),
        ('TangentialMode[nx_, ny_, Lx_, Ly_, c_]', lambda nx, ny, Lx, Ly, c: c/2*sqrt((nx/Lx)**2 + (ny/Ly)**2)),
        ('ObliqueMode[nx_, ny_, nz_, Lx_, Ly_, Lz_, c_]', lambda nx, ny, nz, Lx, Ly, Lz, c: c/2*sqrt((nx/Lx)**2 + (ny/Ly)**2 + (nz/Lz)**2)),
        ('ModalDensity[V_, c_, f_]', lambda V, c, f: 4*pi*V*f**2/c**3),
        ('SchroederFrequency[T60_, V_]', lambda T60, V: 2000*sqrt(T60/V)),
        
        # ===== PSYCHOACOUSTICS (35 functions) =====
        ('LoudnessLevel[SPL_, f_]', lambda SPL, f: SPL),  # Phons, simplified
        ('Loudness[phons_]', lambda phons: 2**((phons - 40)/10)),  # Sones
        ('EqualLoudnessContour[f_]', lambda f: sp.Symbol('SPL')),
        ('FletcherMunsonCurve[f_]', lambda f: sp.Symbol('threshold')),
        ('ThresholdOfHearing[f_]', lambda f: 20*log(f/1000, 10) - 10 if f > 1000 else 0),
        ('ThresholdOfPain[]', lambda: 130),  # dB SPL
        ('DynamicRange[threshold_, ceiling_]', lambda th, ceil: ceil - th),
        ('MaskingLevel[masker_, signal_]', lambda m, s: m - s),
        ('SimultaneousMasking[f_masker_, f_signal_]', lambda fm, fs: abs(fm - fs)),
        ('TemporalMasking[pre_post_]', lambda pp: pp),
        ('CriticalBand[f_]', lambda f: 25 + 75*(1 + 1.4*(f/1000)**2)**0.69),  # Hz, Zwicker
        ('BarkScale[f_]', lambda f: 13*sp.atan(0.00076*f) + 3.5*sp.atan((f/7500)**2)),
        ('ERBScale[f_]', lambda f: 21.4*log(1 + 0.00437*f, 10)),  # Equivalent rectangular bandwidth
        ('CochlearFilterBandwidth[f_]', lambda f: 24.7*(4.37*f/1000 + 1)),
        ('PitchPerception[f_]', lambda f: f),
        ('MissingFundamental[harmonics_]', lambda h: sp.gcd(*h)),
        ('VirtualPitch[partials_]', lambda p: sp.gcd(*p)),
        ('PitchShift[semitones_]', lambda s: 2**(s/12)),
        ('MusicalPitch[f_]', lambda f: 69 + 12*log(f/440, 2)),  # MIDI note number
        ('FrequencyFromMIDI[note_]', lambda n: 440*2**((n - 69)/12)),
        ('JustNoticeDifferenceFrequency[f_]', lambda f: 0.003*f),  # Weber's law
        ('JNDIntensity[I_]', lambda I: 0.1*I),
        ('TimbreDifference[spectrum1_, spectrum2_]', lambda s1, s2: sum(abs(s1i - s2i) for s1i, s2i in zip(s1, s2))),
        ('SpectralCentroid[f_, A_]', lambda f, A: sum(fi*Ai for fi, Ai in zip(f, A))/sum(A)),
        ('SpectralSpread[f_, A_, centroid_]', lambda f, A, c: sqrt(sum(Ai*(fi - c)**2 for fi, Ai in zip(f, A))/sum(A))),
        ('SpectralFlatness[spectrum_]', lambda s: sp.exp(sum(log(si) for si in s)/len(s)) / (sum(s)/len(s))),
        ('SpectralRolloff[spectrum_]', lambda s: sp.Symbol('f_rolloff')),
        ('ZeroCrossingRate[signal_]', lambda sig: sum(1 for i in range(len(sig)-1) if sig[i]*sig[i+1] < 0)/len(sig)),
        ('AttackTime[envelope_]', lambda env: sp.Symbol('t_attack')),
        ('DecayTime[envelope_]', lambda env: sp.Symbol('t_decay')),
        ('SustainLevel[envelope_]', lambda env: sp.Symbol('level')),
        ('ReleaseTime[envelope_]', lambda env: sp.Symbol('t_release')),
        ('PrecedenceEffect[delay_]', lambda delay: delay < 0.04),  # Haas effect
        ('CocktailPartyEffect[SNR_]', lambda SNR: SNR > -6),
        ('BinauralBeatFrequency[f_left_, f_right_]', lambda fl, fr: abs(fl - fr)),
        
        # ===== AUDIO SIGNAL PROCESSING (30 functions) =====
        ('TimeStretch[signal_, factor_]', lambda sig, f: sig),  # Placeholder
        ('PitchShiftSignal[signal_, semitones_]', lambda sig, s: sig),
        ('AutoTune[signal_, scale_]', lambda sig, sc: sig),
        ('VocoderAnalysis[signal_, carriers_]', lambda sig, car: car),
        ('PhaseVocoder[STFT_, time_stretch_, pitch_shift_]', lambda S, ts, ps: S),
        ('GranularSynthesis[grains_]', lambda g: sum(g)),
        ('WavetableSynthesis[table_, index_]', lambda tab, idx: tab[int(idx) % len(tab)]),
        ('FMSynthesis[carrier_, modulator_, index_]', lambda fc, fm, idx: sin(2*pi*fc*sp.Symbol('t') + idx*sin(2*pi*fm*sp.Symbol('t')))),
        ('AMSynthesis[carrier_, modulator_]', lambda fc, fm: (1 + fm)*fc),
        ('SubtractiveSynthesis[oscillator_, filter_]', lambda osc, filt: filt*osc),
        ('AdditiveSynthesis[partials_]', lambda p: sum(p)),
        ('SampleAndHold[signal_, rate_]', lambda sig, rate: sig),
        ('DownsamplingAudio[signal_, factor_]', lambda sig, f: sig[::int(f)]),
        ('UpsamplingAudio[signal_, factor_]', lambda sig, f: sig),
        ('ResamplingAudio[signal_, ratio_]', lambda sig, r: sig),
        ('BandlimitedInterpolation[signal_]', lambda sig: sig),
        ('WindowingFunction[type_, N_]', lambda typ, N: [1]*int(N)),  # Placeholder
        ('HannWindow[n_, N_]', lambda n, N: 0.5*(1 - cos(2*pi*n/N))),
        ('HammingWindow[n_, N_]', lambda n, N: 0.54 - 0.46*cos(2*pi*n/N)),
        ('BlackmanWindow[n_, N_]', lambda n, N: 0.42 - 0.5*cos(2*pi*n/N) + 0.08*cos(4*pi*n/N)),
        ('KaiserWindow[n_, N_, beta_]', lambda n, N, beta: sp.Symbol('I0')(beta*sqrt(1 - (2*n/N - 1)**2))/sp.Symbol('I0')(beta)),
        ('SpectralLeakage[window_]', lambda w: w),
        ('ScallopingLoss[bin_offset_]', lambda offset: sin(pi*offset)/(pi*offset) if offset != 0 else 1),
        ('OverlapAddMethod[windows_]', lambda w: sum(w)),
        ('ShortTimeFourierTransform[signal_, window_, hop_]', lambda sig, win, hop: sp.Symbol('STFT')),
        ('MelFilterbank[f_, N_mels_]', lambda f, Nmels: [0]*int(Nmels)),
        ('MFCCCoefficients[signal_]', lambda sig: [0]*13),
        ('ChromaFeatures[signal_]', lambda sig: [0]*12),
        ('TonalCentroid[chroma_]', lambda chr: (sp.Symbol('x'), sp.Symbol('y'))),
        ('SpectralContrast[spectrum_]', lambda s: sp.Symbol('contrast')),
        
        # ===== NOISE & VIBRATION (20 functions) =====
        ('WhiteNoise[duration_]', lambda d: "noise"),
        ('PinkNoise[duration_]', lambda d: "1/f noise"),
        ('BrownNoise[duration_]', lambda d: "1/f^2 noise"),
        ('BandlimitedNoise[f_low_, f_high_]', lambda flow, fhigh: "noise"),
        ('NoiseFloor[signal_]', lambda sig: min(sig)),
        ('SignalToNoiseRatio[signal_power_, noise_power_]', lambda Ps, Pn: 10*log(Ps/Pn, 10)),
        ('DynamicRangeAudio[peak_, noise_floor_]', lambda peak, nf: 20*log(peak/nf, 10)),
        ('CrestFactor[peak_, RMS_]', lambda peak, rms: 20*log(peak/rms, 10)),
        ('TotalHarmonicDistortion[harmonics_, fundamental_]', lambda h, f: sqrt(sum(hi**2 for hi in h))/f*100),
        ('IntermodulationDistortion[f1_, f2_]', lambda f1, f2: abs(f1 - f2)),
        ('VibrationLevel[a_, a_ref_]', lambda a, aref: 20*log(a/aref, 10)),
        ('VibrationDose[a_, t_]', lambda a, t: a*sqrt(t)),
        ('RMSVibration[signal_]', lambda sig: sqrt(sum(si**2 for si in sig)/len(sig))),
        ('PeakVibration[signal_]', lambda sig: max(abs(si) for si in sig)),
        ('VibrationalEnergy[m_, omega_, A_]', lambda m, omega, A: sp.Rational(1,2)*m*omega**2*A**2),
        ('DampingRatioVibration[zeta_]', lambda zeta: zeta),
        ('QFactorVibration[omega_n_, zeta_]', lambda omegan, zeta: 1/(2*zeta)),
        ('TransmissibilityVibration[omega_, omega_n_, zeta_]', lambda omega, omegan, zeta: sqrt((1 + (2*zeta*omega/omegan)**2)/((1 - (omega/omegan)**2)**2 + (2*zeta*omega/omegan)**2))),
        ('VibrationIsolationEfficiency[T_]', lambda T: (1 - T)*100),
        ('ShockResponseSpectrum[acceleration_, time_]', lambda a, t: sp.Symbol('SRS')),
    ]


def register():
    """Register all extended30 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""
MikoshiLang Extended Functions - Set 21
Climate science, meteorology, and atmospheric physics
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi, sin, cos

def get_rules():
    """Extended set 21: Climate and atmospheric science (150 functions)"""
    T, P, rho, z = symbols('T P rho z')
    
    return [
        # ===== ATMOSPHERIC THERMODYNAMICS (40 functions) =====
        ('IdealGasLaw[P_, V_, n_, R_, T_]', lambda P, V, n, R, T: P*V - n*R*T),
        ('BarometricFormula[P0_, M_, g_, R_, T_, h_]', lambda P0, M, g, R, T, h: P0*exp(-M*g*h/(R*T))),
        ('HypsometricEquation[T_, P1_, P2_]', lambda T, P1, P2: sp.Symbol('R')*T/sp.Symbol('g')*log(P1/P2)),
        ('PotentialTemperature[T_, P_, P0_]', lambda T, P, P0: T*(P0/P)**sp.Rational(2,7)),
        ('VirtualTemperature[T_, q_]', lambda T, q: T*(1 + 0.608*q)),
        ('EquivalentTemperature[T_, q_]', lambda T, q: T + sp.Symbol('L_v')*q/sp.Symbol('c_p')),
        ('WetBulbTemperature[T_, RH_]', lambda T, RH: T*atan2(0.151977*sqrt(RH + 8.313659), 1) + atan2(T + RH, 1) - atan2(RH - 1.676331, 1) + 0.00391838*RH**1.5*atan2(0.023101*RH, 1) - 4.686035),
        ('DewPoint[T_, RH_]', lambda T, RH: T - (100 - RH)/5),  # Simplified
        ('VaporPressure[T_]', lambda T: 6.112*exp(17.67*T/(T + 243.5))),  # Magnus formula
        ('SaturationVaporPressure[T_]', lambda T: 6.112*exp(17.67*T/(T + 243.5))),
        ('RelativeHumidity[e_, es_]', lambda e, es: e/es * 100),
        ('SpecificHumidity[e_, P_]', lambda e, P: 0.622*e/(P - 0.378*e)),
        ('MixingRatio[e_, P_]', lambda e, P: 0.622*e/(P - e)),
        ('AbsoluteHumidity[e_, T_]', lambda e, T: e/(sp.Symbol('R_v')*T)),
        ('LapseRateD ry[g_, cp_]', lambda g, cp: g/cp),
        ('LapseRateMoist[gamma_d_, L_, e_s_]', lambda gd, L, es: gd*(1 + L*es/(sp.Symbol('R')*sp.Symbol('T')**2))/(1 + L**2*es/(sp.Symbol('c_p')*sp.Symbol('R_v')*sp.Symbol('T')**2))),
        ('AdiabaticCooling[gamma_, delta_z_]', lambda gamma, dz: gamma*dz),
        ('LiftedCondensationLevel[T_, Td_]', lambda T, Td: 125*(T - Td)),  # Meters
        ('LevelOfFreeConvection[T_parcel_, T_env_]', lambda Tp, Te: Tp > Te),
        ('ConvectiveAvailablePotentialEnergy[T_parcel_, T_env_, z_]', lambda Tp, Te, z: sp.Symbol('g')*sp.integrate((Tp - Te)/Te, z)),
        ('ConvectiveInhibition[T_parcel_, T_env_, z_]', lambda Tp, Te, z: -sp.Symbol('g')*sp.integrate((Tp - Te)/Te, z)),
        ('BruntVaisalaFrequency[g_, T_, dT_dz_]', lambda g, T, dTdz: sqrt(g/T*dTdz)),
        ('RichardsonNumberAtmos[g_, T_, dT_dz_, du_dz_]', lambda g, T, dTdz, dudz: g/T*dTdz/dudz**2),
        ('StaticStability[dtheta_dz_]', lambda dthetadz: dthetadz > 0),
        ('ConditionalInstability[gamma_, gamma_m_]', lambda gamma, gammam: gamma > gammam),
        ('AbsoluteInstability[gamma_, gamma_d_]', lambda gamma, gammad: gamma > gammad),
        ('EquivalentPotentialTemperature[theta_, q_]', lambda theta, q: theta*exp(sp.Symbol('L_v')*q/(sp.Symbol('c_p')*sp.Symbol('T')))),
        ('CloudBase[LCL_]', lambda LCL: LCL),
        ('CloudTop[T_, gamma_]', lambda T, gamma: T/gamma),
        ('PrecipitableWater[q_, P_]', lambda q, P: sp.integrate(q*P/sp.Symbol('g'), P)),
        ('LatentHeatVaporization[T_]', lambda T: 2.5e6 - 2400*T),  # J/kg
        ('LatentHeatFusion[]', lambda: 3.34e5),  # J/kg
        ('LatentHeatSublimation[]', lambda: 2.83e6),  # J/kg
        ('ClaudiusClapeyron[L_, T_]', lambda L, T: L/(sp.Symbol('R_v')*T**2)),
        ('BoltonFormula[T_, RH_]', lambda T, RH: T - (100 - RH)/5),
        ('TetensFormula[T_]', lambda T: 6.112*exp(17.27*T/(T + 237.3))),
        ('GoffGratchEquation[T_]', lambda T: 6.112*exp(17.67*T/(T + 243.5))),
        ('WexlerEquation[T_]', lambda T: exp(-5800.2206/T + 1.3914993 - 0.048640239*T)),
        ('MagnusFormula[T_]', lambda T: 6.112*exp(17.67*T/(T + 243.5))),
        ('AugustRocheMagnus[T_]', lambda T: 6.1094*exp(17.625*T/(T + 243.04))),
        
        # ===== RADIATION & ENERGY BALANCE (30 functions) =====
        ('SolarConstant[]', lambda: 1361),  # W/m²
        ('SolarZenithAngle[lat_, decl_, hour_]', lambda lat, decl, h: sp.acos(sin(lat)*sin(decl) + cos(lat)*cos(decl)*cos(h))),
        ('SolarDeclination[day_]', lambda d: 23.45*sin(2*pi*(d + 284)/365)),
        ('DayLength[lat_, decl_]', lambda lat, decl: 24/pi*sp.acos(-tan(lat)*tan(decl))),
        ('Insolation[S0_, theta_]', lambda S0, theta: S0*cos(theta)),
        ('TopOfAtmosphereRadiation[S0_, distance_factor_]', lambda S0, df: S0*df),
        ('AlbedoSurface[reflected_, incident_]', lambda refl, inc: refl/inc),
        ('PlanetaryAlbedo[]', lambda: 0.30),  # Earth average
        ('GreenhouseEffect[T_surface_, T_effective_]', lambda Ts, Te: Ts - Te),
        ('BlackbodyRadiation[T_]', lambda T: sp.Symbol('sigma')*T**4),
        ('WienDisplacementLaw[T_]', lambda T: 2898/T),  # μm
        ('StefanBoltzmann[T_]', lambda T: 5.67e-8*T**4),
        ('RayleighScattering[lambda_]', lambda lam: lam**(-4)),
        ('MieScattering[r_, lambda_]', lambda r, lam: r/lam),
        ('OpticalDepth[tau_]', lambda tau: tau),
        ('TransmittanceAtmospheric[tau_]', lambda tau: exp(-tau)),
        ('BeerLambertLaw[I0_, k_, x_]', lambda I0, k, x: I0*exp(-k*x)),
        ('AbsorptionCoefficient[optical_depth_, path_]', lambda tau, x: tau/x),
        ('ScatteringCoefficient[tau_scat_, path_]', lambda taus, x: taus/x),
        ('SingleScatteringAlbedo[tau_scat_, tau_total_]', lambda taus, taut: taus/taut),
        ('AsymmetryParameter[g_]', lambda g: g),  # -1 to 1
        ('AerosolOpticalDepth[tau_]', lambda tau: tau),
        ('CloudOpticalDepth[LWP_, r_eff_]', lambda LWP, reff: 1.5*LWP/reff),
        ('LongwaveRadiation[T_, epsilon_]', lambda T, eps: eps*sp.Symbol('sigma')*T**4),
        ('NetRadiation[SW_down_, SW_up_, LW_down_, LW_up_]', lambda SWd, SWu, LWd, LWu: SWd - SWu + LWd - LWu),
        ('RadiativeForcing[delta_F_]', lambda dF: dF),
        ('CO2RadiativeForcing[C_, C0_]', lambda C, C0: 5.35*log(C/C0)),
        ('ClimatesensitivityParameter[lambda_]', lambda lam: lam),  # °C/(W/m²)
        ('EquilibriumClimateSensitivity[lambda_, forcing_]', lambda lam, F: lam*F),
        ('TransientClimateResponse[warming_]', lambda w: w),
        
        # ===== ATMOSPHERIC DYNAMICS (30 functions) =====
        ('GeostrophicWind[f_, dp_dy_]', lambda f, dpdy: -dpdy/(sp.Symbol('rho')*f)),
        ('AgeostrophicWind[v_total_, v_geostrophic_]', lambda vt, vg: vt - vg),
        ('ThermalWind[f_, dT_dy_, z_]', lambda f, dTdy, z: -sp.Symbol('g')/f*dTdy/sp.Symbol('T')*z),
        ('CoriolisParameter[lat_]', lambda lat: 2*sp.Symbol('Omega')*sin(lat)),
        ('RossbyNumber[U_, L_, f_]', lambda U, L, f: U/(L*f)),
        ('EkmanLayer[nu_, f_]', lambda nu, f: sqrt(2*nu/f)),
        ('EkmanSpiral[z_, f_, nu_]', lambda z, f, nu: exp(-z*sqrt(f/(2*nu)))),
        ('PressureGradientForce[dp_dx_, rho_]', lambda dpdx, rho: -dpdx/rho),
        ('GradientWind[f_, R_, dp_dr_]', lambda f, R, dpdr: -R*f + sqrt(R**2*f**2 + R*dpdr/sp.Symbol('rho'))),
        ('CyclostrophicWind[R_, dp_dr_]', lambda R, dpdr: sqrt(R*dpdr/sp.Symbol('rho'))),
        ('VorticityRelative[du_dy_, dv_dx_]', lambda dudy, dvdx: dvdx - dudy),
        ('VorticityAbsolute[zeta_rel_, f_]', lambda zetar, f: zetar + f),
        ('VorticityPotential[zeta_abs_, h_]', lambda zetaa, h: (zetaa + sp.Symbol('f'))/h),
        ('Divergence[du_dx_, dv_dy_]', lambda dudx, dvdy: dudx + dvdy),
        ('Deformation[du_dx_, dv_dy_, du_dy_, dv_dx_]', lambda dudx, dvdy, dudy, dvdx: sqrt((dudx - dvdy)**2 + (dudy + dvdx)**2)),
        ('VorticityAdvection[v_, grad_zeta_]', lambda v, gradz: -v*gradz),
        ('TemperatureAdvection[v_, grad_T_]', lambda v, gradT: -v*gradT),
        ('OmegaEquation[omega_]', lambda omega: omega),  # Vertical velocity
        ('QuasiGeostrophicVorticity[f_, omega_]', lambda f, omega: sp.diff(omega, sp.Symbol('p'))),
        ('RossbyWave[beta_, k_, l_]', lambda beta, k, l: -beta*k/(k**2 + l**2)),
        ('BaroclinicInstability[f_, N_, dT_dy_]', lambda f, N, dTdy: dTdy > 0),
        ('BarotropicInstability[du_dy_]', lambda dudy: dudy != 0),
        ('ErtelPotentialVorticity[zeta_, theta_]', lambda zeta, theta: (zeta + sp.Symbol('f'))*sp.diff(theta, sp.Symbol('p'))),
        ('IsentropicPV[theta_, u_, v_]', lambda theta, u, v: sp.Symbol('g')*(sp.Symbol('f') + sp.diff(v, sp.Symbol('x')) - sp.diff(u, sp.Symbol('y')))*sp.diff(theta, sp.Symbol('p'))),
        ('MountainWaves[N_, U_, h_]', lambda N, U, h: N*h/U),
        ('GravityWaveDispersion[N_, k_]', lambda N, k: N/k),
        ('KelvinWave[c_, y_, R_]', lambda c, y, R: exp(-y**2/(2*R**2))),
        ('MixedRossbyGravity[beta_, c_]', lambda beta, c: sqrt(beta*c)),
        ('EquatorialBetaPlane[beta_]', lambda beta: beta),
        ('HadleyCirculation[T_equator_, T_pole_]', lambda Teq, Tpole: Teq - Tpole),
        
        # ===== CLOUD & PRECIPITATION PHYSICS (25 functions) =====
        ('DropletGrowth[r_, S_]', lambda r, S: S*sp.Symbol('D_v')/r),
        ('KohlerCurve[r_, S_]', lambda r, S: S - 1),
        ('CriticalSupersaturation[r_]', lambda r: sqrt(4*sp.Symbol('A')**3/(27*r**3))),
        ('CriticalRadius[S_]', lambda S: sqrt(3*sp.Symbol('A')/S)),
        ('CondensationRate[S_, N_]', lambda S, N: S*N*sp.Symbol('D_v')),
        ('CollisionEfficiency[r1_, r2_]', lambda r1, r2: (r1 + r2)**2/(4*r1*r2)),
        ('CoalescenceEfficiency[E_collision_]', lambda Ec: Ec),
        ('TerminalVelocityDrop[r_]', lambda r: sp.Symbol('g')*r**2/(18*sp.Symbol('mu'))),  # Stokes
        ('RaindropSizeDistribution[N0_, lambda_, D_]', lambda N0, lam, D: N0*exp(-lam*D)),
        ('MarshallPalmer[R_]', lambda R: 8000*exp(-41*sp.Symbol('D')*R**(-0.21))),
        ('ReflectivityFactor[Z_]', lambda Z: Z),  # mm⁶/m³
        ('RadarReflectivity[Z_]', lambda Z: 10*log(Z, 10)),  # dBZ
        ('ZRRelation[Z_]', lambda Z: (Z/200)**sp.Rational(1,1.6)),  # Rainfall rate
        ('IceNucleation[T_]', lambda T: T < 273.15),
        ('HomogeneousFreezing[T_]', lambda T: T < 235),
        ('HeterogeneousFreezing[T_, IN_]', lambda T, IN: T < 273.15 and IN > 0),
        ('DepositionGrowth[S_ice_, D_v_]', lambda Si, Dv: Si*Dv),
        ('RimingRate[LWC_, v_]', lambda LWC, v: LWC*v*sp.Symbol('E')),
        ('AggregationRate[N_, v_]', lambda N, v: N**2*v*sp.Symbol('E')),
        ('SnowflakeMass[D_]', lambda D: sp.Symbol('a')*D**sp.Symbol('b')),
        ('GraupelDensity[rho_]', lambda rho: rho),  # 0.1-0.9 g/cm³
        ('HailGrowth[LWC_, v_, D_]', lambda LWC, v, D: LWC*v*pi*D**2/4),
        ('LiquidWaterContent[rho_L_, r_]', lambda rhoL, r: sp.Rational(4,3)*pi*rhoL*r**3*sp.Symbol('N')),
        ('CloudWaterPath[LWC_, H_]', lambda LWC, H: LWC*H),
        ('IceWaterPath[IWC_, H_]', lambda IWC, H: IWC*H),
        
        # ===== CLIMATE INDICES & PHENOMENA (25 functions) =====
        ('ElNinoIndex[SST_]', lambda SST: SST),  # Niño3.4 anomaly
        ('SouthernOscillationIndex[P_Tahiti_, P_Darwin_]', lambda PT, PD: PT - PD),
        ('NorthAtlanticOscillation[P_Iceland_, P_Azores_]', lambda PI, PA: PI - PA),
        ('ArcticOscillation[SLP_]', lambda SLP: SLP),  # First EOF of SLP
        ('PacificDecadalOscillation[SST_]', lambda SST: SST),  # First PC
        ('AtlanticMultidecadalOscillation[SST_]', lambda SST: SST),
        ('IndianOceanDipole[SST_west_, SST_east_]', lambda SSTw, SSTe: SSTw - SSTe),
        ('MaddenJulianOscillation[RMM1_, RMM2_]', lambda RMM1, RMM2: sqrt(RMM1**2 + RMM2**2)),
        ('QuasiBiennialOscillation[U_]', lambda U: U),  # Zonal wind at 30 hPa
        ('AnnularMode[EOF_]', lambda EOF: EOF),
        ('TeleconnectionPattern[corr_field_]', lambda corr: corr),
        ('WalkerCirculation[SST_gradient_]', lambda dSST: dSST),
        ('MonsoonIndex[precip_]', lambda P: P),
        ('DroughtIndex[P_, PET_]', lambda P, PET: P - PET),
        ('StandardizedPrecipitationIndex[P_]', lambda P: (P - sp.Symbol('mean'))/sp.Symbol('std')),
        ('PalmerDroughtIndex[P_, PET_, soil_]', lambda P, PET, soil: P - PET + soil),
        ('HeatWaveIndex[T_, threshold_, duration_]', lambda T, th, dur: sp.Symbol('HeatWaveIndex')),
        ('FreezingDegreedays[T_]', lambda T: sum(max(0, 0 - Ti) for Ti in T)),
        ('GrowingDegreeDays[T_, T_base_]', lambda T, Tb: sum(max(0, Ti - Tb) for Ti in T)),
        ('VorticityIndex[vort_]', lambda vort: vort),
        ('BlockingIndex[Z500_]', lambda Z: Z),
        ('JetStreamIndex[u_max_]', lambda umax: umax),
        ('StormTrackActivity[variance_]', lambda var: var),
        ('StratosphericSuddenWarming[T_strat_]', lambda Tstrat: Tstrat > sp.Symbol('threshold')),
        ('PolarVortexStrength[u_60N_]', lambda u: u),
    ]


def register():
    """Register all extended21 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())
